import numpy as np
from shapely.geometry import MultiLineString, LineString, box, Polygon, MultiPolygon, GeometryCollection
from shapely.ops import linemerge, polygonize, cascaded_union
import geopandas as gpd
from typing import Union
from collections import defaultdict
try:
    import pyvoronoi
except ImportError:
    pyvoronoi = None

GenericLine = Union[LineString, MultiLineString]


class PyVoronoiHelper:
    """Add-on for the pyvoronoi (boost voronoi) tool. It computes the voronoi cells within a bounding box."""

    def __init__(self, points: list, segments: list, bounding_box_coords: list, scaling_factor=100000.0):
        if pyvoronoi is None:
            raise ImportError("Impossible to use Voronoi utils, `pyvoronoi` package not found.")
        self.pv = pyvoronoi.Pyvoronoi(scaling_factor)
        for p in points:
            self.pv.AddPoint(p)
        for s in segments:
            self.pv.AddSegment(s)
        self.pv.Construct()
        self.discretization_tolerance = 10000 / scaling_factor
        self.bounding_box_coords = bounding_box_coords

    def get_cells_as_gdf(self) -> gpd.GeoDataFrame:
        """Returns the voronoi cells in `geodataframe` with a column named `id` referencing the index of the associated
         input geometry."""
        gdf = gpd.GeoDataFrame(columns=["id", "geometry"])
        cells_geometries = self.get_cells_as_polygons()
        gdf["geometry"] = list(cells_geometries.values())
        gdf["id"] = list(cells_geometries.keys())
        return gdf

    def get_cells_as_polygons(self) -> dict:
        """Return the voronoi cells as polygons trimmed with the bounding box."""
        diagonal_length = np.linalg.norm(np.array(self.bounding_box_coords[0]) - np.array(self.bounding_box_coords[1]))
        cells_coordinates = self.get_cells_coordiates(eta=diagonal_length,
                                                      discretization_tolerance=self.discretization_tolerance)
        bounding_box = box(self.bounding_box_coords[0][0], self.bounding_box_coords[0][1],
                           self.bounding_box_coords[1][0], self.bounding_box_coords[1][1])
        cells_as_polygons = dict()
        for i, coords in cells_coordinates.items():
            if len(coords) > 2:
                polygon = Polygon(coords)
                if not polygon.is_valid:
                    polygon = self.repair_polygon(polygon)
                trimmed_polygon = polygon.intersection(bounding_box)
                cells_as_polygons[i] = trimmed_polygon
        return cells_as_polygons

    @staticmethod
    def repair_polygon(polygon: Union[Polygon, MultiPolygon]) -> Union[Polygon, MultiPolygon]:
        """Repair an invalid polygon. It works in most cases but it has no guarantee of success."""
        bowtie_repaired_polygon = PyVoronoiHelper.repair_bowtie_polygon(polygon)
        if not bowtie_repaired_polygon.is_valid:
            return polygon.buffer(0.0)
        else:
            return bowtie_repaired_polygon

    @staticmethod
    def repair_bowtie_polygon(polygon: Union[Polygon, MultiPolygon]) -> MultiPolygon:
        """Repair an invalid polygon for the 'bowtie' case."""
        p_ext = polygon.exterior
        self_intersection = p_ext.intersection(p_ext)
        mp = MultiPolygon(polygonize(self_intersection))
        return mp

    def get_cells_coordiates(self, eta=1.0, discretization_tolerance=0.05) -> dict:
        """"Parse the results of ``pyvoronoi`` to compute the voronoi cells coordinates. The infinite ridges are
        projected at a ``eta`` distance in the ridge direction.

        Parameters
        ----------
        eta : float
            Distance for infinite ridges projection. (Default value = 1.0)
        discretization_tolerance : float
            Discretization distance for curved edges. (Default value = 0.05)

        Returns
        -------
        dict
            A dictionary mapping the cells ids and their coordinates.

        """
        vertices = self.pv.GetVertices()
        cells = self.pv.GetCells()
        edges = self.pv.GetEdges()
        cells_coordinates = dict()
        for c in cells:
            cell_coords = []
            for e in c.edges:
                edge = edges[e]
                start_vertex = vertices[edge.start] if edge.start != -1 else None
                end_vertex = vertices[edge.end] if edge.end != -1 else None
                if edge.start == -1 or edge.end == -1:
                    self.clip_infinite_edge(cell_coords, edge, eta)
                else:
                    if edge.is_linear:
                        self.add_polygon_coordinates(cell_coords, [start_vertex.X, start_vertex.Y])
                        self.add_polygon_coordinates(cell_coords, [end_vertex.X, end_vertex.Y])
                    else:
                        try:
                            coords_to_add = []
                            for p in self.pv.DiscretizeCurvedEdge(e, discretization_tolerance):
                                coords_to_add.append(p)
                            cell_coords.extend(coords_to_add)
                        except pyvoronoi.UnsolvableParabolaEquation:
                            self.add_polygon_coordinates(cell_coords, [start_vertex.X, start_vertex.Y])
                            self.add_polygon_coordinates(cell_coords, [end_vertex.X, end_vertex.Y])
            cells_coordinates[c.cell_identifier] = cell_coords
        return cells_coordinates

    def clip_infinite_edge(self, cell_coords: list, edge: "pyvoronoi.Edge", eta: float):
        """Fill infinite edge coordinate by placing the infinite vertex to a ``eta`` distance of the known vertex."""
        cell = self.pv.GetCell(edge.cell)
        twin_edge = self.pv.GetEdge(edge.twin)
        twin_cell = self.pv.GetCell(twin_edge.cell)
        # Infinite edges could not be created by two segment sites.
        if cell.contains_point and twin_cell.contains_point:
            first_point = self.pv.RetrieveScaledPoint(cell)
            second_point = self.pv.RetrieveScaledPoint(twin_cell)
            origin = np.array([(first_point[0] + second_point[0]) / 2.0,
                               (first_point[1] + second_point[1]) / 2.0])
            ridge_direction = np.array([first_point[1] - second_point[1],
                                        second_point[0] - first_point[0]])
        else:
            if cell.contains_segment:
                origin = np.array(self.pv.RetrieveScaledPoint(twin_cell))
                segment = np.array(self.pv.RetriveScaledSegment(cell))
            else:
                origin = np.array(self.pv.RetrieveScaledPoint(cell))
                segment = np.array(self.pv.RetriveScaledSegment(twin_cell))
            dx = segment[1][0] - segment[0][0]
            dy = segment[1][1] - segment[0][1]
            if (np.linalg.norm(segment[0] - origin) == 0.0) != cell.contains_point:
                ridge_direction = np.array([dy, -dx])
            else:
                ridge_direction = np.array([-dy, dx])
        ridge_direction /= np.linalg.norm(ridge_direction)
        if edge.start == -1:
            ridge_point_projected = origin - ridge_direction * eta
            self.add_polygon_coordinates(cell_coords, ridge_point_projected)
        else:
            start_vertex = self.pv.GetVertex(edge.start)
            self.add_polygon_coordinates(cell_coords, [start_vertex.X, start_vertex.Y])
        if edge.end == -1:
            ridge_point_projected = origin + ridge_direction * eta
            self.add_polygon_coordinates(cell_coords, ridge_point_projected)
        else:
            end_vertex = self.pv.GetVertex(edge.end)
            self.add_polygon_coordinates(cell_coords, [end_vertex.X, end_vertex.Y])


    @staticmethod
    def add_polygon_coordinates(coordinates: list, point: list):
        """Add given point to given coordinates list if is not the equal to the last coordinates."""
        if coordinates:
            last_point = coordinates[-1]
            if last_point[0] == point[0] and last_point[1] == point[1]:
                return
        coordinates.append(point)


def split_linestring_as_simple_linestrings(line: GenericLine) -> list:
    """Split a linestring if it is not simple (i.e. it crosses itself)."""
    if not line.is_simple:
        mls = line.intersection(line)
        if line.geom_type == 'LineString' and mls.geom_type == 'MultiLineString':
            mls = linemerge(mls)
        return list(mls)
    else:
        return [line]

def split_as_simple_segments(lines: list, tol=1e-7) -> defaultdict:
    """Split a list of lines to simple segments (linestring composed by two points). All returned segments do not
    cross themselves except at extremities.

    Parameters
    ----------
    lines : list
        List of lines to split
    tol : float
        Tolerance to test if a line is a sub line of another one. (Default value = 1e-7)

    Returns
    -------
    defaultdict
        A dictionary mapping for each input line index, the list of simple segments.

    """
    split_lines_mapping = defaultdict(list)
    all_split_lines = split_linestring_as_simple_linestrings(MultiLineString(lines))
    j = 0
    sub_line = all_split_lines[j]
    nb_simple_segments = len(all_split_lines)
    end_mapping = False
    for i, line in enumerate(lines):
        while line.buffer(tol, 1).contains(sub_line):
            split_lines_mapping[i].append(sub_line)
            j += 1
            if j >= nb_simple_segments:
                end_mapping = True
                break
            sub_line = all_split_lines[j]
        if end_mapping:
            break
    return split_lines_mapping


def compute_voronoi_cells_from_lines(lines: list, tolerance=1e-7) -> list:
    """Compute the voronoi cells of given generic lines. Input linestrings can be not simple.

    Parameters
    ----------
    lines : list
        List of ``LineString``
    tolerance : float
        Tolerance for the voronoi cells computation (Two points will be considered equal if their
        coordinates are equal when rounded at ``tolerance``). (Default value = 1e-7)

    Returns
    -------
    list
        A list of cells geometries.

    """
    simple_segments_mapping = split_as_simple_segments(lines, tolerance)
    all_segments = [list(s.coords) for i in range(len(lines)) for s in simple_segments_mapping[i]]
    bounds = MultiLineString(lines).bounds
    bb = [[bounds[0], bounds[1]], [bounds[2], bounds[3]]]
    pvh = PyVoronoiHelper([], segments=all_segments, bounding_box_coords=bb, scaling_factor=1 / tolerance)
    gdf = pvh.get_cells_as_gdf()
    gdf = gdf[list(map(lambda i: isinstance(i, Polygon), gdf["geometry"]))].copy()
    gdf["site"] = [pvh.pv.GetCell(c).site for c in gdf["id"]]
    lines_cells = []
    ct = 0
    for i, line in enumerate(lines):
        line_polygons = []
        for s in simple_segments_mapping[i]:
            for p in gdf[gdf["site"] == ct]["geometry"]:
                line_polygons.append(p)
            ct += 1
        merged_polygon = cascaded_union(line_polygons)
        if isinstance(merged_polygon, GeometryCollection):
            if len(merged_polygon) == 0:
                continue
            merged_polygon = MultiPolygon(merged_polygon)
        lines_cells.append(merged_polygon)
    return lines_cells







# -*- coding: utf-8 -*-
import os
import shutil
from nose.tools import assert_is_instance
import unittest
from nose.plugins.attrib import attr
import geonetworkx as gnx
from geonetworkx.testing import assert_graphs_have_same_edges_geometry, ALL_CLASSES, get_random_geograph_subclass
import geonetworkx.testing.utils as gnx_tu


gnx_tu.SEED = 70595
NB_VERTICES = 50


@attr('readwrite')
class TestReadWrite(unittest.TestCase):

    def setUp(self):
        file_dir = os.path.dirname(__file__)
        self.results_dir = os.path.join(file_dir, "datasets/results/")
        if not os.path.exists(self.results_dir):
            os.mkdir(self.results_dir)

    def tearDown(self):
        shutil.rmtree(self.results_dir)

    def test_write_read_gpickle(self):
        for i, graph_type in enumerate(ALL_CLASSES):
            with self.subTest(graph_type=graph_type, SEED=gnx_tu.SEED):
                g = get_random_geograph_subclass(NB_VERTICES, graph_type)
                file_path = os.path.join(self.results_dir, "write_test_%d.gpickle" % i)
                gnx.write_gpickle(g, file_path)
                read_graph = gnx.read_gpickle(file_path)
                assert_is_instance(read_graph, graph_type)
                assert_graphs_have_same_edges_geometry(g, read_graph)

    def test_write_read_graphml(self):
        for i, graph_type in enumerate(ALL_CLASSES):
            with self.subTest(graph_type=graph_type, SEED=gnx_tu.SEED):
                g = get_random_geograph_subclass(NB_VERTICES, graph_type)
                gnx.stringify_nodes(g, copy=False)
                file_path = os.path.join(self.results_dir, "write_test_%d.xml" % i)
                gnx.write_graphml(g, file_path)
                read_graph = gnx.read_graphml(file_path)
                assert_is_instance(read_graph, graph_type)
                assert_graphs_have_same_edges_geometry(g, read_graph)

    def test_write_geofile(self):
        for graph_type in ALL_CLASSES:
            with self.subTest(graph_type=graph_type, SEED=gnx_tu.SEED):
                g = get_random_geograph_subclass(NB_VERTICES, graph_type)
                gnx.write_geofile(g, self.results_dir, driver="GPKG")

    def test_write_read_geofile(self):
        for i, graph_type in enumerate(ALL_CLASSES):
            with self.subTest(graph_type=graph_type, SEED=gnx_tu.SEED):
                g_write = get_random_geograph_subclass(NB_VERTICES, graph_type)
                g_write.name = "test_%d" % i
                gnx.write_geofile(g_write, self.results_dir, driver="GPKG")
                nodes_file_path = os.path.join(self.results_dir, g_write.name + "_nodes.gpkg")
                edges_file_path = os.path.join(self.results_dir, g_write.name + "_edges.gpkg")
                g_read = gnx.read_geofiles(nodes_file_path, edges_file_path,
                                           directed=g_write.is_directed(), multigraph=g_write.is_multigraph())

                from geonetworkx.testing import assert_graphs_have_same_geonodes
                assert_graphs_have_same_geonodes(g_write, g_read, "Written and read graph have different nodes.")
                assert_graphs_have_same_edges_geometry(g_write, g_read,
                                                       "Written and read graph have different edge geometries")
                # This test may not pass for multigraphs with multi-edges because edges may be read in a different order
                # that the are written (to be investigated).



# -*- coding: utf-8 -*-
import geonetworkx.geograph
from geonetworkx.geograph import *

import geonetworkx.geodigraph
from geonetworkx.geodigraph import *

import geonetworkx.geomultigraph
from geonetworkx.geomultigraph import *

import geonetworkx.geomultidigraph
from geonetworkx.geomultidigraph import *

import geonetworkx.geometry_operations
from geonetworkx.geometry_operations import *

import geonetworkx.readwrite
from geonetworkx.readwrite import *

import geonetworkx.geograph
from geonetworkx.simplify import *

import geonetworkx.utils
from geonetworkx.utils import *

import geonetworkx.settings
from geonetworkx.settings import *

import geonetworkx.tools
from geonetworkx.tools import *

import geonetworkx.generators
from geonetworkx.generators import *



import os
import networkx as nx
import geonetworkx as gnx
import geopandas as gpd


def get_copenhagen_street_net():
    """Reads dataset geojson and return the corresponding geograph."""
    dir_path = os.path.dirname(os.path.realpath(__file__))
    nodes_path = os.path.join(dir_path, "copenhagen_streets_net_nodes.geojson")
    edges_path = os.path.join(dir_path, "copenhagen_streets_net_edges.geojson")
    return gnx.read_geofiles(nodes_path, edges_path, directed=True, multigraph=True)


def get_copenhagen_ferry_net():
    """Reads dataset geojson and return the corresponding geograph."""
    dir_path = os.path.dirname(os.path.realpath(__file__))
    nodes_path = os.path.join(dir_path, "copenhagen_ferry_net_nodes.geojson")
    edges_path = os.path.join(dir_path, "copenhagen_ferry_net_edges.geojson")
    return gnx.read_geofiles(nodes_path, edges_path, directed=True, multigraph=True)


def get_copenhagen_buildings():
    """Reads GeoDataFrame from geojson dataset"""
    dir_path = os.path.dirname(os.path.realpath(__file__))
    return gpd.read_file(os.path.join(dir_path, "copenhagen_buildings.geojson"))


def get_grenoble_streets_500():
    """Reads dataset geojson and return the corresponding geograph."""
    dir_path = os.path.dirname(os.path.realpath(__file__))
    nodes_path = os.path.join(dir_path, "grenoble_streets_500_nodes.geojson")
    edges_path = os.path.join(dir_path, "grenoble_streets_500_edges.geojson")
    return gnx.read_geofiles(nodes_path, edges_path, directed=False, multigraph=True)


def get_grenoble_streets_200():
    """Reads dataset geojson and return the corresponding geograph."""
    dir_path = os.path.dirname(os.path.realpath(__file__))
    nodes_path = os.path.join(dir_path, "grenoble_streets_200_nodes.geojson")
    edges_path = os.path.join(dir_path, "grenoble_streets_200_edges.geojson")
    return gnx.read_geofiles(nodes_path, edges_path, directed=False, multigraph=True)


def get_grenoble_streets_isere():
    """Reads dataset geojson and return the corresponding geograph."""
    dir_path = os.path.dirname(os.path.realpath(__file__))
    nodes_path = os.path.join(dir_path, "grenoble_isere_nodes.geojson")
    edges_path = os.path.join(dir_path, "grenoble_isere_edges.geojson")
    return gnx.read_geofiles(nodes_path, edges_path, directed=True, multigraph=True)


def import_paris_subway_dataset():
    try:
        import osmnx as ox
    except ImportError:
        raise RuntimeError("OSMNX package is required for import data from OSM")
    # Set additional info to retrieve
    ox.settings.useful_tags_node.append("name")
    ox.settings.useful_tags_path.append("name")
    # Download raw data with osmnx
    g = ox.graph_from_address("Paris, France", distance=7000, infrastructure='relation["route"="subway"]',
                              clean_periphery=False, simplify=False, custom_filter="")

    # Transform to a geograph
    gg = gnx.read_geograph_with_coordinates_attributes(g)
    gg.name = "Paris_subway"

    # Supplement graph data
    # Add geometry on all edges
    gnx.fill_edges_missing_geometry_attributes(gg)

    # Simplify: remove dead ends and two degree nodes that are not a subway station
    def is_not_station(n):
        return gg.nodes[n].get("name", None) is None

    all_edges_names = nx.get_edge_attributes(gg, "name")
    gnx.remove_dead_ends(gg, node_filter=is_not_station, only_strict=False)
    edges_mapping = gnx.two_degree_node_merge(gg, node_filter=is_not_station)
    # Supplement data in the added edges with the most present value in the merged edges.
    for new_edge, removed_edges in edges_mapping.items():
        available_names = [all_edges_names[e] for e in removed_edges if e in all_edges_names]
        if available_names:
            gg.edges[new_edge]["name"] = max(set(available_names), key=available_names.count)

    # Save geograph
    dir_path = os.path.dirname(os.path.realpath(__file__))
    gnx.write_gpickle(gg, os.path.join(dir_path, "paris_subway.gpickle"))



# -*- coding: utf-8 -*-
# default CRS to set when creating graphs
WGS84_CRS = {'init': 'epsg:4326'}
USED_CRS = WGS84_CRS
DEFAULT_CRS = None

# default discretization parameter for extended projection methods
DISCRETIZATION_TOLERANCE = 1e-4

# Graph default spatial keys
NODES_GEOMETRY_DEFAULT_KEY = "geometry"
EDGES_GEOMETRY_DEFAULT_KEY = "geometry"


# Intersection prefix for spatial merges
INTERSECTION_PREFIX = "intersect_"
ORIGINAL_EDGE_KEY = "original_edge"
BOUNDARY_NODE_PREFIX = "boundary_"

# read/write convention
EDGE_FIRST_NODE_COLUMN_NAME = "u"
EDGE_SECOND_NODE_COLUMN_NAME = "v"
NODE_ID_COLUMN_NAME = "id"
KNOWN_FILES_EXTENSION = {'GeoJSON': '.geojson', 'GPKG': '.gpkg', 'ESRI Shapefile': '.shp'}



# -*- coding: utf-8 -*-
import networkx as nx
from .geograph import GeoGraph
import geonetworkx as gnx


class GeoDiGraph(GeoGraph, nx.DiGraph):
    """Base class for directed geographic graphs.
    
    Because edges are directed, it supposes that the edges lines are well-ordered. Namely, that the first point of the
    line matches with the coordinates of the first vertex of the edge (or is at least close) and vice versa with the
    last point of the line and the second. If this is not the case, the method ``order_well_lines`` can be useful to
    make sure of that.
    """

    def to_nx_class(self):
        """ """
        return nx.DiGraph

    def to_undirected(self, reciprocal=False, as_view=False):
        """Return an undirected copy of the graph (see ``networkx.DiGraph.to_undirected``)."""
        if as_view:
            return nx.DiGraph.to_undirected(self, reciprocal, as_view)
        else:
            graph_class = self.to_undirected_class()
            undirected_graph = nx.DiGraph.to_undirected(self, reciprocal, as_view)
            return graph_class(undirected_graph)

    def to_undirected_class(self):
        """Returns the class to use for empty undirected copies (see ``networkx.DiGraph.to_undirected_class``)."""
        return gnx.GeoGraph

    def to_directed(self, as_view=False):
        """Return a directed representation of the graph (see ``networkx.DiGraph.to_directed``)."""
        if as_view:
            return nx.DiGraph.to_directed(self, as_view)
        else:
            graph_class = self.to_directed_class()
            directed_graph = nx.DiGraph.to_directed(self, as_view)
            return graph_class(directed_graph)

    def to_directed_class(self):
        """Returns the class to use for empty directed copies (see ``networkx.DiGraph.to_directed_class``)."""
        return GeoDiGraph

    """
    TODO
    def reverse():
        call nx.DiGraph reverse
        Reverse all linestrings
    """



# -*- coding: utf-8 -*-
import networkx as nx
from .geomultigraph import GeoMultiGraph
from .geodigraph import GeoDiGraph
import geonetworkx as gnx


class GeoMultiDiGraph(GeoMultiGraph, GeoDiGraph, nx.MultiDiGraph):
    """A directed geographic graph class that can store multiedges."""

    def to_nx_class(self):
        """Return the closest networkx class (in the inheritance graph)."""
        return nx.MultiDiGraph

    def to_undirected(self, reciprocal=False, as_view=False):
        """Return an undirected copy of the graph (see ``networkx.MultiDiGraph.to_undirected``)."""
        if as_view:
            return nx.MultiDiGraph.to_undirected(self, reciprocal, as_view)
        else:
            graph_class = self.to_undirected_class()
            undirected_graph = nx.MultiDiGraph.to_undirected(self, reciprocal, as_view)
            return graph_class(undirected_graph)

    def to_undirected_class(self):
        """Returns the class to use for empty undirected copies (see ``networkx.MultiDiGraph.to_undirected_class``)."""
        return gnx.GeoMultiGraph

    def to_directed(self, as_view=False):
        """Return a directed representation of the graph (see ``networkx.MultiDiGraph.to_directed``)."""
        if as_view:
            return nx.MultiDiGraph.to_directed(self, as_view)
        else:
            graph_class = self.to_directed_class()
            directed_graph = nx.MultiDiGraph.to_directed(self, as_view)
            return graph_class(directed_graph)

    def to_directed_class(self):
        """Returns the class to use for empty directed copies (see ``networkx.MultiDiGraph.to_directed_class``)."""
        return GeoMultiDiGraph



# -*- coding: utf-8 -*-
import networkx as nx
from geonetworkx.geograph import GeoGraph
from geonetworkx import settings
from geonetworkx.geometry_operations import split_line
from geonetworkx.utils.geograph_utils import get_line_start


def _get_ego_boundaries(graph: GeoGraph, ego_graph: GeoGraph, n, radius: float, distance=None) -> tuple:
    """Retrieve all information to build an extended ego-graph. See ``gnx.extended_ego_graph`` for more info."""
    if graph.is_multigraph():
        out_edges_options = {"keys": True}
    else:
        out_edges_options = {}
    boundary_nodes = []
    boundary_edges = dict()
    for u in ego_graph.nodes:
        node_edges = graph.edges(u, **out_edges_options)
        for e in node_edges:
            v = e[1]
            if v in ego_graph.nodes:
                continue
            edge_data = graph.edges[e]
            if graph.edges_geometry_key not in edge_data:
                continue
            # The distance d(n, u) has already been computed in the networkx ego graph method,
            # but for the sake of code simplicity, we are doing it again here.
            d_n_u = nx.dijkstra_path_length(graph, n, u, weight=distance)
            edge_distance = edge_data.get(distance, 1)
            distance_margin = radius - d_n_u
            interpolation_factor = distance_margin / edge_distance
            edge_geometry = edge_data[graph.edges_geometry_key]
            edge_length = edge_geometry.length
            if get_line_start(graph, e, edge_geometry) == u:
                interpolation_distance = interpolation_factor * edge_length
                inside_line_part_index = 0
            else:
                interpolation_distance = (1.0 - interpolation_factor) * edge_length
                inside_line_part_index = 1
            if interpolation_distance <= 0 or interpolation_distance >= edge_length:
                continue
            b_node_point = edge_geometry.interpolate(interpolation_distance)
            b_node_name = settings.BOUNDARY_NODE_PREFIX + str(u) + '_' + str(v)
            boundary_nodes.append((b_node_name, {graph.nodes_geometry_key: b_node_point}))
            split_edge = split_line(edge_geometry, interpolation_distance)
            inside_edge_data = {graph.edges_geometry_key: split_edge[inside_line_part_index],
                                distance: distance_margin}
            outside_edge_data = {graph.edges_geometry_key: split_edge[1 - inside_line_part_index],
                                 distance: edge_distance - distance_margin}
            boundary_edges[e] = [(u, b_node_name, inside_edge_data), (b_node_name, v, outside_edge_data)]
    return boundary_nodes, boundary_edges


def extended_ego_graph(graph: GeoGraph, n, radius=1, center=True, undirected=False, distance=None) -> GeoGraph:
    """Returns induced subgraph of neighbors centered at node n within a given radius extended by interpolated nodes on
    boundary edges.

    Note that the returned graph is not a subgraph of the input graph because it will have boundary nodes in addition.
    It means that a node is added on each edge leaving the ego-graph to represent the furthest reachable point on
    this edge. The boundary node is added at given node using a linear interpolation. A boundary node :math:`b` will
    be added on the edge :math:`(u, v)` if :math:`d(n, u) < radius < d(n, v)`. The boundary will be placed along the
    edge geometry at the following distance:

    .. math::
        d(u, b) =  \\frac{\\text{radius} - d(n, u)}{d(u, v)}

    Furthermore, the attribute ``distance`` is filled with the value :math:`d(u, b)`.

    Parameters
    ----------
    graph : GeoGraph, GeoDiGraph, GeoMultiGraph or GeoMultiDiGraph
        A Geograph or subclass
    n :
        A single source node
    radius : float or int
        Include all neighbors of distance<=radius from n. (Default value = 1)
    center : bool
        If False, do not include center node in graph (Default value = True)
    undirected : bool
        If True use both in- and out-neighbors of directed graphs. (Default value = False)
    distance : str
        Use specified edge data key as distance.  For example, setting distance='weight' will use the edge
        weight to measure the distance from the node n. (Default value = None)

    Returns
    -------
    GeoGraph, GeoDiGraph, GeoMultiGraph or GeoMultiDiGraph
        The resulting graph with node, edge, and graph attributes copied.

    See Also
    --------
    add_ego_boundary_nodes

    """
    ego_graph = nx.ego_graph(graph, n, radius, center=True, undirected=undirected, distance=distance)
    if undirected:
        working_graph = graph.to_undirected()
    else:
        working_graph = graph
    # Retrieve boundary edge and cut geometries
    boundary_nodes, boundary_edges = _get_ego_boundaries(working_graph, ego_graph, n, radius, distance)
    ego_graph.add_nodes_from(boundary_nodes)
    edges_to_add = []
    for e, ego_edges in boundary_edges.items():
        edges_to_add.append(ego_edges[0])
    ego_graph.add_edges_from(edges_to_add)
    if not center:
        ego_graph.remove_node(n)
    return ego_graph


def add_ego_boundary_nodes(graph: GeoGraph, n, radius, distance, undirected=False):
    """Modify the given graph to add boundary nodes at exact radius distance. An edge :math:`(u, v)` is a boundary edge
    if :math:`(u, v)` if :math:`d(n, u) < radius < d(n, v)`. A boundary node is added on the edge to represent the ego-
    graph limit. See ``gnx.extended_ego_graph`` for more info.

    Parameters
    ----------
    graph : GeoGraph, GeoDiGraph, GeoMultiGraph or GeoMultiDiGraph
        Input graph to modify
    n :
        A single source node
    radius : float or int
        Include all neighbors of distance<=radius from n.
    distance : str
        Use specified edge data key as distance. For example, setting distance='weight' will use the edge
        weight to measure the distance from the node n.
    undirected : bool
        If True use both in- and out-neighbors of directed graphs. (Default value = False)

    See Also
    --------
    extended_ego_graph

    """
    ego_graph = nx.ego_graph(graph, n, radius, center=True, undirected=undirected, distance=distance)
    if undirected:
        working_graph = graph.to_undirected()
    else:
        working_graph = graph
    # Retrieve boundary edge and cut geometries
    boundary_nodes, boundary_edges = _get_ego_boundaries(working_graph, ego_graph, n, radius, distance)
    graph.add_nodes_from(boundary_nodes)
    edges_to_add = []
    for e, ego_edges in boundary_edges.items():
        edges_to_add.extend(ego_edges)
        graph.remove_edge(*e)
    graph.add_edges_from(edges_to_add)



# -*- coding: utf-8 -*-
""" This example shows how a two geographs can be merged in a routing use case. Here, a street network and a subway
network is combined to find the multi-modal shortest path between a source and a target."""
import os
import geopandas as gpd
import networkx as nx
import geonetworkx as gnx
import osmnx as ox


# Constants: speed in each network
bicycle_speed = 10 * 1000 / 3600  # 10 km/h -> 2.8 m/s

# Download and set up the street network (main streets_graph only)
streets_graph = ox.graph_from_address("Rennes, France", distance=2500,
                                      infrastructure='way["highway"~"primary|secondary|tertiary"]')
streets_graph = gnx.read_geograph_with_coordinates_attributes(streets_graph)
streets_graph.name = "Rennes_streets"

# Getting the bicycle stations
datasets_path = os.path.join(os.path.dirname(os.path.realpath(__file__)), "../tests/datasets/")
bicycle_stations = gpd.read_file(os.path.join(datasets_path, "rennes_bicycle_stations_velo_star.geojson"))

# Merging the stations to the street network
gnx.spatial_points_merge(streets_graph, bicycle_stations, inplace=True)

# Setting times on the edges
gnx.fill_edges_missing_geometry_attributes(streets_graph)
gnx.fill_length_attribute(streets_graph, attribute_name="length", only_missing=True)
streets_lengths = nx.get_edge_attributes(streets_graph, "length")
nx.set_edge_attributes(streets_graph, {e: l / bicycle_speed for e, l in streets_lengths.items()}, "time")


# Computing all shortest path between all stations with a 15min cutoff
cutoff = 15 * 60  # in seconds
bicycle_stations_nodes = list(bicycle_stations.index)

# Creating a graph connecting all pairs of reachable stations within 15 min
bicycle_stations_graph = gnx.GeoDiGraph()
for n in bicycle_stations_nodes:
    bicycle_stations_graph.add_node(n, geometry=bicycle_stations.at[n, "geometry"])

for n in bicycle_stations_nodes:
    distances = nx.single_source_dijkstra_path_length(streets_graph,
                                                      source=n,
                                                      weight="time",
                                                      cutoff=cutoff)
    for u, d in distances.items():
        if u != n and u in bicycle_stations_nodes:
            bicycle_stations_graph.add_edge(n, u, time=d)

bicycle_stations_graph.name = "bicycle_station_graph"
gnx.write_geofile(bicycle_stations_graph, datasets_path)






# -*- coding: utf-8 -*-
import numpy as np
from geonetworkx.geograph import GeoGraph
from geonetworkx.utils import get_line_start
import geonetworkx as gnx
from shapely.geometry import Polygon, LineString, MultiPolygon
from shapely.ops import cascaded_union
import math
from typing import Union
from scipy.spatial import Delaunay
import geopandas as gpd


GenericPolygon = Union[Polygon, MultiPolygon]


def get_edges_voronoi_cells(graph: GeoGraph, tolerance=1e-7) -> gpd.GeoSeries:
    """Return edge voronoi cells as `GeoSeries`."""
    edge_as_lines = graph.get_edges_as_line_series()
    lines = list(edge_as_lines)
    edge_cells = gnx.compute_voronoi_cells_from_lines(lines, tolerance)
    edge_cells_as_series = gpd.GeoSeries({e: cell for e, cell in zip(edge_as_lines.index, edge_cells)})
    edge_cells_as_series.crs = graph.crs
    return edge_cells_as_series


def get_segment_boundary_buffer_polygon(segment_coords: Union[list, np.array],
                                        radius: float, residual_radius: float) -> Polygon:
    """Return a segment boundary polygon using given radius. It represents all reachable points from the first
    extremity of the segment. The returned polygon is a trapeze. See ``boundary_edge_buffer``."""
    segment_direction = [segment_coords[1][0] - segment_coords[0][0], segment_coords[1][1] - segment_coords[0][1]]
    orthogonal_dir = np.array([- segment_direction[1], segment_direction[0]])
    orthogonal_dir /= np.linalg.norm(orthogonal_dir)
    top_points = [segment_coords[0] + orthogonal_dir * radius, segment_coords[1] + orthogonal_dir * residual_radius]
    bottom_points = [segment_coords[0] - orthogonal_dir * radius, segment_coords[1] - orthogonal_dir * residual_radius]
    return Polygon(top_points + list(reversed(bottom_points)))


def get_point_boundary_buffer_polygon(point_coords: list, radius: float,
                                      segment_direction: list, resolution=16) -> Polygon:
    """Returns a half-disk centered on the given point, with the given radius and having the boundary edge orthogonal to
    the given segment direction. See ``boundary_edge_buffer``."""
    # Segment angle with system coordinates
    phi = math.acos(segment_direction[0] / np.linalg.norm(segment_direction))
    if segment_direction[1] < 0:
        phi *= -1.0
    # Discretization angle
    theta = math.pi / float(resolution)
    # Start angle
    angle = phi - math.pi / 2.0
    coords = []
    for i in range(resolution + 1):
        coords.append(point_coords + radius * np.array([math.cos(angle), math.sin(angle)]))
        angle -= theta
    return Polygon(coords)


def boundary_edge_buffer(line: LineString) -> GenericPolygon:
    """Return the edge buffer polygon on the oriented line. This represented the area where all points are reachable
    starting from the line first extremity and using the closest edge projection rule."""
    radius = line.length
    residual_radius = radius
    boundary_polygons = []
    for i in range(len(line.coords) - 1):
        segment_coords = np.array([line.coords[i], line.coords[i + 1]])
        residual_radius -= gnx.euclidian_distance_coordinates(segment_coords[0], segment_coords[1])
        boundary_polygon = get_segment_boundary_buffer_polygon(segment_coords, radius, residual_radius)
        boundary_polygons.append(boundary_polygon)
        segment_direction = segment_coords[1] - segment_coords[0]
        boundary_polygons.append(get_point_boundary_buffer_polygon(segment_coords[0], radius, segment_direction))
        radius = residual_radius
    return cascaded_union(boundary_polygons)


def isochrone_polygon(graph: GeoGraph, source, limit, weight="length", tolerance=1e-7) -> GenericPolygon:
    """Return a polygon approximating the isochrone set in the geograph.

    Parameters
    ----------
    graph : Geograph
        Graph representing possible routes.
    source :
        Source node from where distance is computed
    limit : float or int
        Isochrone limit (e.g. 100 meters, 5 minutes, depending on ``weight`` unit).
    weight : str
        Weight attribute on edges to compute distances (edge weights should be non-negative). (Default value = "length")
    tolerance : float
        Tolerance to compute Voronoi cells. (Default value = 1e-7)

    Returns
    -------
    Polygon or MultiPolygon
        A polygon representing all reachable points within the given limit from the source node.

    """
    working_graph = graph.copy()
    # Compute the ego-graph
    gnx.add_ego_boundary_nodes(working_graph, source, limit, distance=weight)
    ego_graph = gnx.extended_ego_graph(working_graph, source, limit, distance=weight)
    # Compute edges voronoi cells
    edge_voronoi_cells = get_edges_voronoi_cells(working_graph, tolerance)
    # Set ego-graph edges cells
    isochrone_polygons = []
    for e, edge in enumerate(edge_voronoi_cells.index):
        if ego_graph.has_edge(*edge):
            p = edge_voronoi_cells[e]
            if not graph.has_edge(*edge):  # For boundary edges
                boundary_line = working_graph.edges[edge][working_graph.edges_geometry_key]
                if get_line_start(working_graph, edge, boundary_line) != edge[0]:
                    boundary_line = LineString(reversed(boundary_line.coords))
                edge_buffer_pol = boundary_edge_buffer(boundary_line)
                isochrone_polygons.append(p.intersection(edge_buffer_pol))
            else:
                isochrone_polygons.append(p)
    # Merge as isochrone polygon
    final_polygon = cascaded_union(isochrone_polygons)
    final_polygon = final_polygon.buffer(tolerance)
    return final_polygon


def get_alpha_shape_polygon(points: list, quantile: float) -> GenericPolygon:
    """Return the alpha-shape polygon formed by the given points. Alpha parameter is determined using a quantile of
    circumradius of Delaunay triangles.

    Parameters
    ----------
    points : list
        List of input points (2D)
    quantile : float
        Quantile on circumradius to determine alpha (100 returns the convex hull,
        0 returns an empty polygon). ``0 <= quantile <= 100``.

    Returns
    -------
    Polygon or MultiPolygon
        The polygon formed by all triangles having a circumradius inferior or equal to :math:`1/\\alpha`.
        
        Note that this does not return the exhaustive alpha-shape for low quantiles, the minimum spanning tree LineString
        should be added to the returned polygon.
        This is adapted from `Sean Gillies code <https://sgillies.net/2012/10/13/the-fading-shape-of-alpha.html>`_.

    """
    points = np.asarray(points)
    tri = Delaunay(points)
    polygons = []
    # loop over triangles:
    # ia, ib, ic = indices of corner points of the triangle
    circum_radius = []
    for ia, ib, ic in tri.vertices:
        pa = points[ia]
        pb = points[ib]
        pc = points[ic]
        # Lengths of sides of triangle
        a = gnx.euclidian_distance_coordinates(pa, pb)
        b = gnx.euclidian_distance_coordinates(pb, pc)
        c = gnx.euclidian_distance_coordinates(pc, pa)
        # Semiperimeter of triangle
        s = (a + b + c) / 2.0
        # Area of triangle by Heron's formula
        area = math.sqrt(np.clip(s * (s - a) * (s - b) * (s - c), 0.0, float("inf")))
        if area <= 0.0:
            continue
        circum_r = a * b * c / (4.0 * area)
        circum_radius.append(circum_r)
        polygons.append(Polygon([pa, pb, pc]))
    inv_alpha = np.percentile(circum_radius, quantile)
    filtered_polygons = [p for i, p in enumerate(polygons) if circum_radius[i] <= inv_alpha]
    return cascaded_union(filtered_polygons)


def isochrone_polygon_with_alpha_shape(graph: GeoGraph, source, limit,
                                       weight="length",
                                       alpha_quantile=99.0,
                                       remove_holes=True,
                                       tolerance=1e-7) -> GenericPolygon:
    """Returns an approximation of the isochrone polygon using an alpha-shape of the Shortest Path Tree.

    Parameters
    ----------
    graph : GeoGraph
        GeoGraph to browse
    source :
        Source node from where distance is computed
    limit : float or int
        Isochrone limit (e.g. 100 meters, 5 minutes, depending on ``weight`` unit).
    weight : str
        Weight attribute on edges to compute distances (edge weights should be non-negative). (Default value = "length")
    alpha_quantile : float
        Quantile on circumradius to determine alpha (100 returns the convex hull,
        0 returns an empty polygon). ``0 <= quantile <= 100``. (Default value = 99.0)
    remove_holes : bool
        If ``True`` remove holes in the returned polygon. (Default value = True)
    tolerance : float
        Buffering tolerance on polygon for rendering (Default value = 1e-7)

    Returns
    -------
    Polygon or MultiPolygon
        A polygon approximating the isochrone.

    """
    # Compute the ego-graph
    ego_graph = gnx.extended_ego_graph(graph, source, limit, distance=weight)
    edge_as_lines = ego_graph.get_edges_as_line_series()
    discretized_lines, _ = gnx.discretize_lines(edge_as_lines)
    alpha_shape = get_alpha_shape_polygon(discretized_lines, alpha_quantile)
    alpha_shape = alpha_shape.buffer(tolerance)
    if remove_holes:
        if isinstance(alpha_shape, MultiPolygon):
            alpha_shape = MultiPolygon([Polygon(sub_p.exterior) for sub_p in alpha_shape])
        else:
            alpha_shape = Polygon(alpha_shape.exterior)
    return alpha_shape



# -*- coding: utf-8 -*-
import math
import numpy as np
import networkx as nx
from shapely.geometry import Point, LineString, MultiPoint
import geopy.distance
import pyproj
from geonetworkx.geometry_operations import coordinates_almost_equal, insert_point_in_line
from geonetworkx.geograph import GeoGraph
import geonetworkx.settings as settings
from typing import Iterable
try:
    import srtm
except ImportError:
    srtm = None


def get_crs_as_str(crs) -> str:
    """Return the given CRS as string ``pyproj.Proj`` methods."""
    proj = pyproj.Proj(crs)
    return proj.definition_string()


def is_null_crs(crs) -> bool:
    """Test for null crs values."""
    if crs is None:
        return True
    if crs == dict():
        return True
    if crs == '':
        return True
    return False


def crs_equals(crs1, crs2) -> bool:
    """Compare CRS using ``pyproj.Proj`` objects."""
    if is_null_crs(crs1) or is_null_crs(crs1):
        return False
    return get_crs_as_str(crs1) == get_crs_as_str(crs2)


def compute_vincenty(p1, p2) -> float:
    """Returns the vincenty distance in meters given points with the format (longitude, latitude) in the WGS84
    crs."""
    return geopy.distance.distance((p1[1], p1[0]), (p2[1], p2[0])).meters


def compute_vincenty_from_points(p1: Point, p2: Point) -> float:
    return compute_vincenty([p1.x, p1.y], [p2.x, p2.y])


def approx_map_unit_factor(points_coordinates, tolerance=1e-7) -> float:
    """Compute a linear approximation of the map unit factor for 1 meter. Works only for the WGS84 CRS."""
    centroid = np.array(points_coordinates)
    lower_bound = centroid
    initial_gap = tolerance
    while compute_vincenty(centroid, centroid + initial_gap) < 1.0:
        initial_gap *= 2
    upper_bound = centroid + initial_gap
    unit_point = (lower_bound + upper_bound) / 2
    distance = compute_vincenty(centroid, unit_point)
    iteration = 0
    max_iterations = 10000
    while np.abs(distance - 1.0) > tolerance:
        if distance > 1.0:
            upper_bound = unit_point
        else:
            lower_bound = unit_point
        unit_point = (lower_bound + upper_bound) / 2
        distance = compute_vincenty(centroid, unit_point)
        iteration += 1
        if iteration > max_iterations:
            break
    return np.linalg.norm(centroid - unit_point)


def measure_line_distance(line: LineString) -> float:
    """Measure the length of a shapely LineString object using the vincenty distance.

    Parameters
    ----------
    line : LineString
        Linestring to measure. Coordinates have to be (in the WGS-84 ellipsoid model)

    Returns
    -------
    float
        distance in meters of the linestring.

    """
    coords = line.coords
    if len(coords) < 2:
        return 0.0
    u = coords[0]
    total_distance = 0.0
    for i in range(1, len(coords)):
        v = coords[i]
        total_distance += compute_vincenty(u, v)
        u = v
    return total_distance


def get_new_node_unique_name(graph: nx.Graph, name: str):
    """Return a new unique node name from an initial node name. A counter suffix is added at the end if the node name is
    already used.

    Parameters
    ----------
    graph : nx.Graph
        A given graph
    name : str
        A initial node name

    Returns
    -------
    str
        A unique name not in ``graph.nodes()``.

    """
    if name not in graph.nodes():
        return name
    else:
        ct = 1
        unique_name = name
        while unique_name in graph.nodes():
            ct += 1
            unique_name = "%s_%d" % (str(name), ct)
        return unique_name


def euclidian_distance_coordinates(c1: Iterable, c2: Iterable) -> float:
    """Return the euclidian distance between the two sets of coordinates."""
    return math.sqrt(sum(((i - j) ** 2 for i, j in zip(c1, c2))))


def euclidian_distance(p1: Point, p2: Point) -> float:
    """Return the euclidian distance between the two points

    Parameters
    ----------
    p1 : Point
        The first shapely Point
    p2 : Point
        The second shapely Point

    Returns
    -------
    float
        The euclidian distance

    """
    return euclidian_distance_coordinates((p1.x, p1.y), (p2.x, p2.y))


def fill_edges_missing_geometry_attributes(graph: GeoGraph):
    """Add a geometry attribute to the edges that don't have any. The created geometry is a straight line between the
    two nodes.

    Parameters
    ----------
    graph : GeoGraph
        graph to fill

    """
    edges = dict(graph.edges)
    nodes = dict(graph.nodes)
    for s in edges:
        if graph.edges_geometry_key not in edges[s]:
            if graph.nodes_geometry_key in nodes[s[0]] and graph.nodes_geometry_key in nodes[s[1]]:
                p1 = nodes[s[0]][graph.nodes_geometry_key]
                p2 = nodes[s[1]][graph.nodes_geometry_key]
                graph.edges[s][graph.edges_geometry_key] = LineString([p1, p2])


def fill_length_attribute(graph: GeoGraph, attribute_name="length", only_missing=True):
    """Fill the ``'length'`` attribute of the given networkX Graph. The length is computed in meters using the vincenty
    formula. Method won't be consistent if the graph crs is not WGS84.

    Parameters
    ----------
    graph : GeoGraph
        graph to fill
    attribute_name : str
        The length attribute name to set (Default value = "length")
    only_missing : bool
        Compute the length only if the attribute is missing (Default value = True)
    """
    if not crs_equals(graph.crs, settings.USED_CRS):
        raise ValueError("Impossible to compute distance for graph with different"
                         " crs than '%s'. Graph crs: '%s' " % (str(settings.USED_CRS), str(graph.crs)))
    edges_geometry = nx.get_edge_attributes(graph, graph.edges_geometry_key)
    for e in edges_geometry:
        edge_data = graph.edges[e]
        if (not only_missing) or attribute_name not in edge_data:
            edge_data[attribute_name] = measure_line_distance(edges_geometry[e])


def fill_elevation_attribute(graph: GeoGraph, attribute_name="elevation[m]", only_missing=True):
    """Fill the ``elevation[m]`` attribute on nodes of the given geograph. The elevation is found with the `srtm`
    package. Graph crs has to be WGS84 standard, otherwise elevation data won't be consistent.

    Parameters
    ----------
    graph : GeoGraph
        GeoGraph to modify
    attribute_name : str
        Attribute to fill (Default value = "elevation[m]")
    only_missing : bool
        Get the elevation and set it only if the node attribute is missing. (Default value = True)

    """
    if srtm is None:
        raise ImportError("Impossible to get elevation data, `srtm` package not found.")
    elevation_data = srtm.get_data()
    for n, data in graph.nodes(data=True):
        if (not only_missing) or attribute_name not in data:
            longitude, latitude = tuple(*data[graph.nodes_geometry_key].coords)
            elevation = elevation_data.get_elevation(latitude, longitude)
            if elevation is not None:
                data[attribute_name] = elevation


def join_lines_extremity_to_nodes_coordinates(graph: GeoGraph):
    """Modify the edges geometry attribute so that lines extremities match with nodes coordinates.

    Parameters
    ----------
    graph : GeoGraph
        A geograph to modify

    """
    edges_geometry = nx.get_edge_attributes(graph, graph.edges_geometry_key)
    for e in edges_geometry:
        line = edges_geometry[e]
        to_replace = False
        first_node_coords = graph.get_node_coordinates(e[0])
        if not coordinates_almost_equal(line.coords[0], first_node_coords):
            line = insert_point_in_line(line, first_node_coords, 0)
            to_replace = True
        second_node_coords = graph.get_node_coordinates(e[1])
        if not coordinates_almost_equal(line.coords[-1], second_node_coords):
            line = insert_point_in_line(line, second_node_coords, len(line.coords))
            to_replace = True
        if to_replace:
            graph.edges[e][graph.edges_geometry_key] = line


def get_line_start(graph: GeoGraph, e, line):
    """For a given edge, return the node constituting the line start with a closest node rule."""
    uxy = graph.get_node_coordinates(e[0])
    vxy = graph.get_node_coordinates(e[1])
    first_extremity = line.coords[0]
    last_extremity = line.coords[-1]
    u_e1 = euclidian_distance_coordinates(uxy, first_extremity)
    v_e1 = euclidian_distance_coordinates(vxy, first_extremity)
    u_e2 = euclidian_distance_coordinates(uxy, last_extremity)
    v_e2 = euclidian_distance_coordinates(vxy, last_extremity)
    if u_e1 > v_e1 and u_e2 < v_e2:  # u is closer to e2 and v is closer to e1
        return e[1]
    elif u_e1 < v_e1 and u_e2 < v_e2:  # u is closer to e1 and e2 than v
        if u_e1 > u_e2:
            return e[1]
    elif u_e1 > v_e1 and u_e2 > v_e2:  # v in closer to e1 and e2 than u
        if v_e1 < v_e2:
            return e[1]
    else:
        return e[0]


def get_line_ordered_edge(graph: GeoGraph, e, line):
    """Return the given edge with the first node of the edge representing the first line point and the second node
    the last edge point. The closest node rule is applied."""
    if get_line_start(graph, e, line) != e[0]:
        return (e[1], e[0], *e[2:])
    else:
        return e


def order_well_lines(graph: GeoGraph):
    """Try to order well each geometry attribute of edges so that the first coordinates of the line string are the
    coordinates of the first vertex of the edge. The closest node rule is applied. If the graph is not oriented, the
    modification will be inconsistent (nodes declaration in edges views are not ordered). Euclidian distance is used
    here.

    Parameters
    ----------
    graph : GeoGraph
        Graph on which to apply the ordering step. Modification is inplace.

    """
    line_strings = nx.get_edge_attributes(graph, graph.edges_geometry_key)
    for e, line in line_strings.items():
        if get_line_start(graph, e, line) != e[0]:
            graph.edges[e][graph.edges_geometry_key] = LineString(reversed(line.coords))


def stringify_nodes(graph: nx.Graph, copy=True):
    """Modify the graph node names into strings."""
    nx.relabel_nodes(graph, {n: str(n) for n in graph.nodes}, copy)


def is_nan(val) -> bool:
    return val is np.nan or val != val


def rename_nodes_attribute(graph: nx.Graph, old_name, new_name):
    """Rename nodes attribute defined by its old name to a new name."""
    for n, d in graph.nodes(data=True):
        if old_name in d:
            d[new_name] = d.pop(old_name)


def rename_edges_attribute(graph: nx.Graph, old_name, new_name):
    """Rename edges attribute defined by its old name to a new name."""
    for u, v, d in graph.edges(data=True):
        if old_name in d:
            d[new_name] = d.pop(old_name)


def hard_write_spatial_keys(graph: GeoGraph):
    """Write spatial keys in the graph attribute, so that if the default keys are used, they are propagated for
    special operations (e.g. composing graphs)."""
    for k, v in graph.get_spatial_keys().items():
        setattr(graph, k, v)


def compose(G: GeoGraph, H: GeoGraph) -> GeoGraph:
    """Return a new graph of G composed with H. Makes sure the returned graph is consistent with respect to the spatial
    keys. (See ``nx.compose``). According to the priority rule in networkX, attributes from ``H`` take precedent over
    attributes from G."""
    hard_write_spatial_keys(G)
    hard_write_spatial_keys(H)
    R = nx.compose(G, H)
    if G.nodes_geometry_key != H.nodes_geometry_key:
        rename_nodes_attribute(R, G.nodes_geometry_key, H.nodes_geometry_key)
    if G.edges_geometry_key != H.edges_geometry_key:
        rename_edges_attribute(R, G.edges_geometry_key, H.edges_geometry_key)
    return R


def geographical_distance(graph: GeoGraph, node1, node2, method="vincenty") -> float:
    """Return the geographical distance between the two given nodes.

    Parameters
    ----------
    graph : Geograph
        Geograph
    node1 :
        First node label
    node2 :
        Second node label
    method :
        vincenty", "euclidian", "great_circle" (Default value = "vincenty")

    Returns
    -------
    float
        Distance between nodes, unit depends on the method.

    """
    point1 = graph.nodes[node1][graph.nodes_geometry_key]
    point2 = graph.nodes[node2][graph.nodes_geometry_key]
    if method == "vincenty":
        return compute_vincenty_from_points(point1, point2)
    elif method == "euclidian":
        return euclidian_distance(point1, point2)
    else:
        raise ValueError("Unknown distance method: '%s'" % str(method))


def get_graph_bounding_box(graph: GeoGraph):
    """Return the bounding box coordinates of the given GeoGraph. It takes into account nodes and edges geometries."""
    nodes = list(graph.get_nodes_as_point_series())
    x_min, y_min, x_max, y_max = MultiPoint(nodes).bounds
    bb = [[x_min, y_min], [x_max, y_max]]
    edges = graph.get_edges_as_line_series()
    if len(edges) > 0:
        edges_bounds = edges.bounds
        x_e_min, y_e_min = edges_bounds["minx"].min(), edges_bounds["miny"].min()
        x_e_max, y_e_max = edges_bounds["maxx"].max(), edges_bounds["maxy"].max()
        if x_min > x_e_min:
            bb[0][0] = x_e_min
        if y_min > y_e_min:
            bb[0][1] = y_e_min
        if x_max < x_e_max:
            bb[1][0] = x_e_max
        if y_max < y_e_max:
            bb[1][1] = y_e_max
    return bb



# -*- coding: utf-8 -*-
import networkx as nx
from .geograph import GeoGraph
import geonetworkx as gnx


class GeoMultiGraph(GeoGraph, nx.MultiGraph):
    """A undirected geographic graph class that can store multiedges."""

    def to_nx_class(self):
        """Return the closest networkx class (in the inheritance graph)."""
        return nx.MultiGraph

    def to_directed(self, as_view=False):
        """Return a directed representation of the graph (see ``networkx.MultiGraph.to_directed``)."""
        if as_view:
            return nx.MultiGraph.to_directed(self, as_view)
        else:
            graph_class = self.to_directed_class()
            directed_graph = nx.MultiGraph.to_directed(self, as_view)
            return graph_class(directed_graph)

    def to_directed_class(self):
        """Returns the class to use for empty directed copies (see ``networkx.MultiGraph.to_directed_class``)."""
        return gnx.GeoMultiDiGraph

    def to_undirected(self, as_view=False):
        """Return an undirected copy of the graph (see ``networkx.MultiGraph.to_undirected``)."""
        if as_view:
            return nx.MultiGraph.to_undirected(self, as_view)
        else:
            graph_class = self.to_undirected_class()
            undirected_graph = nx.MultiGraph.to_undirected(self, as_view)
            return graph_class(undirected_graph)

    def to_undirected_class(self):
        """Returns the class to use for empty undirected copies (see ``networkx.MultiGraph.to_undirected_class``).."""
        return GeoMultiGraph



# -*- coding: utf-8 -*-
"""Base class for geographic graphs"""
import networkx as nx
import geopandas as gpd
from shapely.geometry import Point, LineString
import geonetworkx as gnx
import geonetworkx.settings as settings


class GeoGraph(nx.Graph):
    """This class extends the ``networkx.Graph`` to represent a graph that have a geographical meaning. Nodes are
    located with their coordinates (x, y) (using ``shapely.geometry.Point`` objects) and edges can be represented with a
    given broken line (using ``shapely.geometry.LineString`` objects). Each graph has its own keys for naming nodes and
    edges geometry (``nodes_geometry_key``, ``edges_geometry_key``). A coordinate reference system (CRS) can be defined
    for a graph and will be used for some methods managing earth coordinates (especially for distances). For now, the
    only supported CRS is the WGS84 standard (EPSG:4326). All nodes must have defined coordinates, otherwise an error
    will be raised  .

    Raises
    ------
    ValueError
        If the all nodes don't have valid coordinates.

    See Also
    --------
    networkx.Graph
    GeoDiGraph
    GeoMultiGraph
    GeoMultiDiGraph

    """
    def __init__(self, incoming_graph_data=None, **attr):
        super(GeoGraph, self).__init__(incoming_graph_data, **attr)
        self.check_nodes_validity()

    def check_nodes_validity(self):
        """Check that all nodes have geometries."""
        for n, node_data in self.nodes(data=True):
            if self.nodes_geometry_key not in node_data:
                raise ValueError("Unable to find geometry for node: '%s'" % str(n))

    @property
    def nodes_geometry_key(self):
        """Attribute name for the edges geometry attributes. This graph attribute appears in the attribute dict
        `G.graph` keyed by the string ``"edges_geometry_key"`` as well as an attribute ``G.nodes_geometry_key``"""
        return self.graph.get('nodes_geometry_key', settings.NODES_GEOMETRY_DEFAULT_KEY)

    @nodes_geometry_key.setter
    def nodes_geometry_key(self, s: str):
        """Sets the node geometry key with the given value"""
        self.graph['nodes_geometry_key'] = s

    @property
    def edges_geometry_key(self):
        """Attribute name for the edges geometry attributes. This graph attribute appears in the attribute dict
        `G.graph` keyed by the string ``"edges_geometry_key"`` as well as an attribute ``G.edges_geometry_key``"""
        return self.graph.get('edges_geometry_key', settings.EDGES_GEOMETRY_DEFAULT_KEY)

    @edges_geometry_key.setter
    def edges_geometry_key(self, s: str):
        """Sets the edges geometry key with the given value"""
        self.graph['edges_geometry_key'] = s

    @property
    def crs(self):
        """Coordinate Reference System of the graph. This graph attribute appears in the attribute dict `G.graph` keyed
        by the string ``"crs"`` as well as an attribute ``G.crs``"""
        return self.graph.get('crs', gnx.DEFAULT_CRS)

    @crs.setter
    def crs(self, c):
        """Sets the current crs"""
        self.graph['crs'] = c

    def get_node_coordinates(self, node_name) -> list:
        """Return the coordinates of the given node.

        Parameters
        ----------
        node_name
            Name of the node on which the coordinates are browsed.

        Returns
        -------
        list
            A two-element list containing (x,y) coordinates of the given node.

        See Also
        --------
        get_nodes_coordinates, get_node_as_point, get_nodes_as_points

        """
        point = self.get_node_as_point(node_name)
        return [point.x, point.y]

    def get_nodes_coordinates(self) -> dict:
        """Return all nodes coordinates within a dictionary.
        
        Returns
        -------
        dict
            Dictionary containing the coordinates of the each node of the graph.

        See Also
        --------
        get_node_coordinates, get_node_as_point, get_nodes_as_points

        """
        return {n: self.get_node_coordinates(n) for n in self.nodes}

    def get_node_as_point(self, node_name):
        """Return a node as a ``shapely.geometry.Point`` object.

        Parameters
        ----------
        node_name :
            Name of the node on which the geometry is browsed.

        Returns
        -------
        shapely.geometry.Point
            The point representing the located node.

        See Also
        --------
        get_node_coordinates, get_node_coordinates, get_nodes_as_points

        """
        node_data = self.nodes[node_name]
        return node_data[self.nodes_geometry_key]

    def get_nodes_as_points(self) -> dict:
        """Return all nodes as ``shapely.geometry.Point`` objects within a dictionary.

        Returns
        -------
        dict
            Dictionary containing the geometry of each node of the graph.

        See Also
        --------
        get_node_coordinates, get_node_coordinates, get_node_as_point

        """
        return {n: self.get_node_as_point(n) for n in self.nodes}

    def get_nodes_as_point_series(self) -> gpd.GeoSeries:
        """Return the nodes as a ``geopandas.GeoSeries`` of ``shapely.geometry.Point``.

        Returns
        -------
        gpd.GeoSeries
            Series containing all nodes geometries. Its CRS is the graph CRS.

        See Also
        --------
        nodes_to_gdf, get_edges_as_line_series

        """
        nodes_as_points = self.get_nodes_as_points()
        point_series = gpd.GeoSeries(nodes_as_points)
        point_series.crs = self.crs
        return point_series

    def get_edges_as_line_series(self) -> gpd.GeoSeries:
        """Return the edges as a ``geopandas.GeoSeries`` of ``shapely.geometry.LineString``.

        Returns
        -------
        gpd.GeoSeries
            Series containing all edges geometries. Its CRS is the graph CRS.

        See Also
        --------
        edges_to_gdf, get_nodes_as_point_series

        """
        lines = nx.get_edge_attributes(self, self.edges_geometry_key)
        line_series = gpd.GeoSeries(lines)
        line_series.crs = self.crs
        return line_series

    def get_spatial_keys(self) -> dict:
        """Return the current graph spatial keys.

        Returns
        -------
        dict
            Dictionary containing spatial keys (nodes and edges geometry keys and crs).
        """
        return {'nodes_geometry_key': self.nodes_geometry_key,
                'edges_geometry_key': self.edges_geometry_key,
                'crs': self.crs}

    def set_nodes_coordinates(self, coordinates: dict):
        """Set nodes coordinates with a given dictionary of coordinates (can be used for a subset of all nodes).

        Parameters
        ----------
        coordinates: dict :
            Dictionary mapping node names and two-element list of coordinates.
        """
        for n, coords in coordinates.items():
            node_data = self.nodes[n]
            node_data[self.nodes_geometry_key] = Point(coords)

    def to_nx_class(self):
        """Return the closest networkx class (in the inheritance graph)."""
        return nx.Graph

    def copy(self, as_view=False):
        """Return a copy of the graph (see ``networkx.Graph.copy``)."""
        nx_graph_class = self.to_nx_class()
        graph = nx_graph_class.copy(self, as_view)
        return self.__class__(graph)

    def to_directed(self, as_view=False):
        """Return a directed representation of the graph (see ``networkx.Graph.to_directed``)."""
        if as_view:
            return nx.Graph.to_directed(self, as_view)
        else:
            graph_class = self.to_directed_class()
            directed_graph = nx.Graph.to_directed(self, as_view)
            return graph_class(directed_graph)

    def to_directed_class(self):
        """Returns the class to use for empty directed copies (see ``networkx.Graph.to_directed_class``)."""
        return gnx.GeoDiGraph

    def to_undirected(self, as_view=False):
        """Return an undirected copy of the graph (see ``networkx.Graph.to_undirected``)."""
        if as_view:
            return nx.Graph.to_undirected(self, as_view)
        else:
            graph_class = self.to_undirected_class()
            undirected_graph = nx.Graph.to_undirected(self, as_view)
            return graph_class(undirected_graph)

    def to_undirected_class(self):
        """Returns the class to use for empty undirected copies (see ``networkx.Graph.to_undirected_class``)."""
        return gnx.GeoGraph

    def to_crs(self, crs=None, epsg=None, inplace=False):
        """Transform nodes and edges geometries to a new coordinate reference system.

        Parameters
        ----------
        crs : dict or str
             Output projection parameters as string or in dictionary form (Default value = None).
        epsg : int
             EPSG code specifying output projection.
        inplace : bool
             If True, the modification is done inplace, otherwise a new graph is created (Default value = False).

        Returns
        -------
        None or GeoGraph
            Nothing is returned if the transformation is inplace, a new GeoGraph is returned otherwise.

        See Also
        --------
        geopandas.GeoSeries.to_crs
        """
        if self.crs is None:
            raise ValueError('Cannot transform naive geometries. Please set a crs on the graph first.')
        if inplace:
            graph = self
        else:
            graph = self.copy()
        # Get new nodes coordinates
        nodes_as_points = graph.get_nodes_as_point_series()
        transformed_nodes = nodes_as_points.to_crs(crs, epsg)
        # Get new edges coordinates
        edges_as_lines = graph.get_edges_as_line_series()
        transformed_edges = edges_as_lines.to_crs(crs, epsg)
        # Operate the transformation
        for n, point in transformed_nodes.iteritems():
            node_data = graph.nodes[n]
            node_data[graph.nodes_geometry_key] = point
        for e, line in transformed_edges.iteritems():
            edge_data = graph.edges[e]
            edge_data[graph.edges_geometry_key] = line
        graph.crs = transformed_nodes.crs
        if not inplace:
            return graph

    def nodes_to_gdf(self) -> gpd.GeoDataFrame:
        """Create a ``geopandas.GeoDataFrame`` from nodes of the current graph. The column representing the geometry is
        named after the current ``nodes_geometry_key`` attribute.

        Returns
        -------
        gpd.GeoDataFrame
            The resulting GeoDataFrame : one row is a node

        See Also
        --------
        get_nodes_as_point_series, edges_to_gdf

        """
        nodes = {node: data for node, data in self.nodes(data=True)}
        gdf_nodes = gpd.GeoDataFrame(nodes).T
        gdf_nodes[settings.NODE_ID_COLUMN_NAME] = gdf_nodes.index
        gdf_nodes.set_geometry(self.nodes_geometry_key, inplace=True)
        gdf_nodes.crs = self.crs
        return gdf_nodes

    def edges_to_gdf(self) -> gpd.GeoDataFrame:
        """Create a ``gpd.GeoDataFrame`` from edges of the current graph. The column representing the geometry is
        named after the current ``edges_geometry_key`` attribute.

        Returns
        -------
        gdf_edges: geopandas.GeoDataFrame
            The resulting GeoDataFrame : one row is an edge

        See Also
        --------
        get_edges_as_line_series, nodes_to_gdf

        """
        # create a list to hold our edges, then loop through each edge in the graph
        edges = []
        for u, v, data in self.edges(data=True):
            # for each edge, add key and all attributes in data dict to the edge_details
            edge_details = {settings.EDGE_FIRST_NODE_COLUMN_NAME: u, settings.EDGE_SECOND_NODE_COLUMN_NAME: v}
            edge_details.update(data)
            # if edge doesn't already have a geometry attribute, create one now
            if self.edges_geometry_key not in data:
                point_u = self.nodes[u][self.nodes_geometry_key]
                point_v = self.nodes[v][self.nodes_geometry_key]
                edge_details[self.edges_geometry_key] = LineString([point_u, point_v])
            edges.append(edge_details)
        # create a GeoDataFrame from the list of edges and set the CRS
        gdf_edges = gpd.GeoDataFrame(edges)
        gdf_edges.set_geometry(self.edges_geometry_key, inplace=True)
        gdf_edges.crs = self.crs
        return gdf_edges

    def add_nodes_from_gdf(self, gdf: gpd.GeoDataFrame, node_index_attr=None):
        """Add nodes with the given `GeoDataFrame` and fill nodes attributes with the geodataframe columns.

        Parameters
        ----------
        gdf :
            GeoDataFrame representing nodes to add (one row for one node).
        node_index_attr :
            Node index attribute for labeling nodes. If ``None``, the dataframe index is used, else
            the given column is used. (Default value = None)

        See Also
        --------
        add_edges_from_gdf

        """
        if not (gnx.is_null_crs(self.crs) or gnx.is_null_crs(gdf.crs) or gnx.crs_equals(gdf.crs, self.crs)):
            gdf = gdf.to_crs(self.crs, inplace=False)
        if node_index_attr is not None:
            gdf = gdf.set_index(node_index_attr, drop=True, inplace=False)
        if gdf._geometry_column_name != self.nodes_geometry_key:
            gdf = gdf.rename(columns={gdf._geometry_column_name: self.nodes_geometry_key}, inplace=False)
            gdf.set_geometry(self.nodes_geometry_key, inplace=True)
        self.add_nodes_from(gdf.iterrows())
        self.check_nodes_validity()

    def add_edges_from_gdf(self, gdf: gpd.GeoDataFrame, edge_first_node_attr=None, edge_second_node_attr=None):
        """Add edges with the given `GeoDataFrame`. If no dataframe columns are specified for first and second node,
        the dataframe index must be a multi-index `(u, v)`.

        Parameters
        ----------
        gdf :
            GeoDataFrame representing edges to add (one row for one edge).
        edge_first_node_attr :
            Edge first node attribute. If ``None``, the dataframe index is used, else the given
            column is used. Must be used with ``edge_second_node_attr``. (Default value = None)
        edge_second_node_attr :
            Edge second node attribute. If ``None``, the dataframe index is used, else the
            given column is used. Must be used with ``edge_first_node_attr``. (Default value = None)

        See Also
        --------
        add_nodes_from_gdf
        """
        if not (gnx.is_null_crs(self.crs) or gnx.is_null_crs(gdf.crs) or gnx.crs_equals(gdf.crs, self.crs)):
            gdf = gdf.to_crs(self.crs, inplace=False)
        if edge_first_node_attr is not None and edge_second_node_attr is not None:
            gdf = gdf.set_index([edge_first_node_attr, edge_second_node_attr], drop=True, inplace=False)
        if gdf._geometry_column_name != self.edges_geometry_key:
            gdf = gdf.rename(columns={gdf._geometry_column_name: self.edges_geometry_key}, inplace=False)
            gdf.set_geometry(self.edges_geometry_key, inplace=True)
        self.add_edges_from((*r[0], r[1]) for r in gdf.iterrows())



# -*- coding: utf-8 -*-
import geonetworkx as gnx
from geonetworkx.tests import datasets


# Read data set: Street graph sample of Grenoble, France.
graph = datasets.get_grenoble_streets_isere()

# Clean data: remove edges orientation, remove self-loops
graph = graph.to_undirected()
gnx.remove_self_loop_edges(graph)

# Supplement graph data: edges geometries, edges length
gnx.fill_edges_missing_geometry_attributes(graph)
gnx.fill_length_attribute(graph, attribute_name="length", only_missing=True)

# Setting isochrone parameters: source, limits
source = 2192254241
limits = [200, 400, 600]  # meters

# Compute isochrone polygon
polygons = []
for l in limits:
    p = gnx.isochrone_polygon(graph, source, l, weight="length")
    polygons.append(p)





# -*- coding: utf-8 -*-
""" This example shows how a two geographs can be merged in a routing use case. Here, a street network and a subway
network is combined to find the multi-modal shortest path between a source and a target."""
import os
import networkx as nx
import geonetworkx as gnx
from geonetworkx.tests import datasets
import osmnx as ox
from shapely.geometry import box

# Constants: speed in each network
bike_speed = 10 * 1000 / 3600  # 10 km/h -> 2.8 m/s
subway_speed = 50 * 1000 / 3600  # 50 km/h -> 13.9 m/s

# Download and set up the street network (this can be long depending on the connection speed)
streets = ox.graph_from_address("Paris, France", distance=2000)
streets = gnx.read_geograph_with_coordinates_attributes(streets)
streets.name = "Paris_streets"
# Setting times on the edges
streets_lengths = nx.get_edge_attributes(streets, "length")
nx.set_edge_attributes(streets, {e: l / bike_speed for e, l in streets_lengths.items()}, "time")

# Getting the subway network
dir_path = os.path.dirname(os.path.realpath(__file__))
input_path = os.path.join(dir_path, "paris_subway.gpickle")
if not os.path.exists(input_path):
    datasets.import_paris_subway_dataset()
subway = gnx.read_gpickle(input_path)
# Setting the times on the edges
gnx.fill_length_attribute(subway)
subway_lengths = nx.get_edge_attributes(subway, "length")
nx.set_edge_attributes(subway, {e: l / subway_speed for e, l in subway_lengths.items()}, "time")
bb = gnx.get_graph_bounding_box(streets)
streets_bb = box(bb[0][0], bb[0][1], bb[1][0], bb[1][1])


def subway_node_is_reachable_from_streets(n):
    """A subway node is reachable from streets if it is a station (i.e. it has a name) and if it is in the bounding box
    of the the street network."""
    if subway.nodes[n].get("name", None) is None:
        return False
    if not streets_bb.contains(subway.nodes[n][subway.nodes_geometry_key]):
        return False
    return True


# Merging the two graphs
streets_and_subway = gnx.spatial_graph_merge(streets, subway, node_filter=subway_node_is_reachable_from_streets)
streets_and_subway.name = "Paris_streets_subway"

# Computing the shortest path
path = nx.shortest_path(streets_and_subway, source=25273868, target=4518793163)
# Saving the optimal path on edges with an attribute
edge_like_path = [(path[i], path[i + 1], 0) for i in range(len(path) - 1)]
for e in edge_like_path:
    streets_and_subway.edges[e]["opt_path"] = 1






# -*- coding: utf-8 -*-
import networkx as nx
from shapely.geometry import Polygon, MultiPolygon, LineString
from geonetworkx import GeoGraph, GeoDiGraph
from geonetworkx.utils import is_nan
from geonetworkx.geometry_operations import merge_two_lines_with_closest_extremities
from typing import Union
from networkx.classes.filters import no_filter


def remove_isolates(graph: nx.Graph) -> int:
    """Removes all isolates nodes in the given graph.

    Parameters
    ----------
    graph : nx.Graph
        A graph on which to remove all isolates

    Returns
    -------
    int
        Number of removed isolates

    """
    isolates = list(nx.isolates(graph))
    graph.remove_nodes_from(isolates)
    return len(isolates)


def remove_self_loop_edges(graph: nx.Graph) -> int:
    """Remove self loop edges on nodes of the given graph.

    Parameters
    ----------
    graph : nx.Graph
        A graph on which to remove all self loops.

    Returns
    -------
    int
        The number of removed self loops

    """
    self_loops_edges = list(nx.selfloop_edges(graph))
    graph.remove_edges_from(self_loops_edges)
    return len(self_loops_edges)


def remove_small_connected_components(graph: nx.Graph, minimum_allowed_size: int) -> int:
    """Remove all connected components having strictly less than ``minimum_allowed_size``.

    Parameters
    ----------
    graph : nx.Graph
        The graph on which to remove connected components
    minimum_allowed_size : int
        The minimum number of nodes where a connected component is kept.

    Returns
    -------
    int
        The number of removed connected components

    """
    connected_components = list(nx.connected_components(graph))
    nb_removed_cc = 0
    for c_ix, cc in enumerate(connected_components):
        if len(cc) < minimum_allowed_size:
            graph.remove_nodes_from(cc)
            nb_removed_cc += 1
        else:
            for n in cc:
                graph.nodes[n]['cc'] = c_ix
    return nb_removed_cc


def trim_graph_with_polygon(graph: GeoGraph, polygon: Union[Polygon, MultiPolygon], copy=False, method="intersects"):
    """Trim a graph with a given polygon. Keep only the nodes and edges that intersect (or are within) the polygon.

    Parameters
    ----------
    graph : GeoGraph, GeoDiGraph, GeoMultiGraph or GeoMultiDiGraph
        A GeoGraph (or subclass)
    polygon : Polygon or MultiPolygon
        A ``shapely.Polygon`` describing the area to keep
    copy : bool
        If ``True``, a deep copy is done and a new graph is returned. (Default value = False)
    method : str
        If set to ``"intersects"``, the ``shapely.intersects`` is used (keeps nodes and edges that
        intersects the polygon). If set to ``"within"``, the ``shapely.within`` is used (keep nodes and edges that are
        strictly into the polygon). (Default value = "intersects")

    Returns
    -------
    None or GeoGraph
        The modified graph if ``copy`` is ``True``.

    """
    if copy:
        used_graph = graph.copy()
    else:
        used_graph = graph
    if method not in ["intersects", "within"]:
        raise ValueError("Unknown method for trimming : '%s'" % str(method))
    nodes_series = used_graph.get_nodes_as_point_series()
    edges_as_series = used_graph.get_edges_as_line_series()
    if method == 'intersects':
        nodes_criteria = ~ nodes_series.intersects(polygon)
        edges_criteria = ~ edges_as_series.intersects(polygon)
    else:
        nodes_criteria = ~ nodes_series.within(polygon)
        edges_criteria = ~ edges_as_series.within(polygon)
    nodes_to_remove = nodes_series[nodes_criteria].index
    edges_to_remove = edges_as_series[edges_criteria].index
    used_graph.remove_nodes_from(nodes_to_remove)
    used_graph.remove_edges_from(edges_to_remove)
    if copy:
        return used_graph


def remove_nan_attributes(graph: nx.Graph, remove_nan=True, remove_none=True, copy=False):
    """Remove the `nan` and `None` values from nodes and edges attributes.

    Parameters
    ----------
    graph : nx.Graph
        Graph (or subclass)
    remove_nan :
        If true, remove the `nan` values (test is ``val is np.nan``) (Default value = True)
    remove_none :
        If true, remove the ``None`` values (test is ``val is None``) (Default value = True)
    copy :
        If True, a copy of the graph is returned, otherwise the graph is modified inplace. (Default value = False)

    Returns
    -------
    None or nx.Graph
        The modified graph if ``copy`` is true.

    """
    if copy:
        used_graph = graph.copy()
    else:
        used_graph = graph

    def trim_data(d):
        keys_to_remove = set()
        for k, val in d.items():
            if remove_none and val is None:
                keys_to_remove.add(k)
            if remove_nan and is_nan(val):
                keys_to_remove.add(k)
        for k in keys_to_remove:
            del d[k]

    for n, data in used_graph.nodes(data=True):
        trim_data(data)
    for u, v, data in used_graph.edges(data=True):
        trim_data(data)
    if copy:
        return used_graph


def get_dead_ends(graph: nx.Graph, node_filter=no_filter, only_strict=False) -> list:
    """Return the list of dead end in the given graph. A dead end is defined as a node having only one neighbor. For
    directed graphs, a strict dead end is a node having a unique predecessor and no successors. A weak dead end is a
    node having a unique predecessor that is also its unique successor.

    Parameters
    ----------
    graph : nx.Graph
        Graph to parse.
    node_filter :
        Evaluates to true if a node can be considered as dead end, false otherwise. (Default value = no_filter)
    only_strict :
        If true, remove only strict dead ends. Used only for directed graphs. (Default value = False)

    Returns
    -------
    list
        List of node name that are dead ends.
    """
    if graph.is_directed():
        dead_ends = []
        for n in graph.nodes():
            if not node_filter(n):
                continue
            nb_predecessors = len(graph.pred[n])
            if nb_predecessors != 1:
                continue
            nb_successors = len(graph.succ[n])
            if nb_successors == 0:
                dead_ends.append(n)
            elif not only_strict and nb_successors == 1:
                pred = next(iter(graph.pred[n]))
                if pred in graph.successors(n):
                    dead_ends.append(n)
        return dead_ends
    else:
        return [n for n in graph.nodes() if node_filter(n) and len(graph.neighbors(n)) == 1]


def remove_dead_ends(graph: nx.Graph, node_filter=no_filter, only_strict=False):
    """Remove dead ends from a given graph. A dead end is defined as a node having only one neighbor. For
    directed graphs, a strict dead end is a node having a unique predecessor and no successors. A weak dead end is a
    node having a unique predecessor that is also its unique successor.

    Parameters
    ----------
    graph : nx.Graph
        Graph to simplify
    node_filter :
        Evaluates to true if a node can be removed, false otherwise. (Default value = no_filter)
    only_strict :
        If true, remove only strict dead ends. Used only for directed graphs. (Default value = False)

    """
    nodes_to_remove = get_dead_ends(graph, node_filter, only_strict)
    while nodes_to_remove:
        graph.remove_nodes_from(nodes_to_remove)
        nodes_to_remove = get_dead_ends(graph, node_filter, only_strict)


def _clean_merge_mapping(edge_mapping: dict, new_edge: tuple, old_edges: list, directed: bool):
    """For the two-degree node merge operation, it cleans the new-old edges mapping dictionary by reporting original
    edges to the newest edge. It makes sure that all edges in the mapping dictionary dict are in the resulting graph.
    """
    for e in old_edges:
        old_edge = None
        if e in edge_mapping.keys():
            old_edge = e
        elif not directed:
            reversed_edge = (e[1], e[0], *e[2:])
            if reversed_edge in edge_mapping.keys():
                old_edge = reversed_edge
        if old_edge is not None:
            if e in edge_mapping[new_edge]:
                edge_mapping[new_edge].remove(e)
            edge_mapping[new_edge].extend(edge_mapping[old_edge])
            del edge_mapping[old_edge]


def two_degree_node_merge_for_directed_graphs(graph: GeoDiGraph, node_filter=no_filter) -> dict:
    """Merge edges that connects two nodes with a unique third node. A potential node to merge `n` must have exactly two
    different neighbors `u` and `v` with one of the following set of edges:
        * `(u, n)` and `(n, v)`
        * `(u, n)`, `(n, u)`, `(n, v)` and `(v, n)`
    
    For the first case, a merging edge `(u, v)` is added. Under the latter, two edges `(u, v)` and `(v, u)` are added.
    The added edges will have a geometry corresponding to concatenation of the two replaced edges. If a replaced edge
    doesn't have a geometry, the added edge will not have a geometry as well. Edges geometries must be well ordered
    (first node must match with line's first extremity), otherwise lines concatenation may not be consistent (see
    ``order_well_lines``).

    Parameters
    ----------
    graph : GeoDiGraph or GeoMultiDiGraph
        Given graph to modify
    node_filter :
        Evaluates to true if a given node can be merged. (Default value = no_filter)

    Returns
    -------
    dict
        Dictionary indicating for each new edge the merged ones.

    """
    def _get_merging_line(g: GeoDiGraph, e1: tuple, e2: tuple) -> Union[LineString, None]:
        first_edge_geometry = g.edges[e1].get(g.edges_geometry_key, None)
        second_edge_geometry = g.edges[e2].get(g.edges_geometry_key, None)
        if first_edge_geometry is not None and second_edge_geometry is not None:
            return LineString(list(first_edge_geometry.coords) + list(second_edge_geometry.coords))
        else:
            return None

    merged_edges = dict()
    nodes = list(graph.nodes)
    for n in nodes:
        if not node_filter(n):
            continue
        in_degree = graph.in_degree(n)
        out_degree = graph.out_degree(n)
        merging_edges = []
        if in_degree == out_degree == 1:
            predecessor = next(iter(graph.pred[n]))
            successor = next(iter(graph.succ[n]))
            if predecessor == successor:
                continue
            if graph.is_multigraph():
                new_edge = (predecessor, successor, graph.new_edge_key(predecessor, successor))
                edges = [(predecessor, n, next(iter(graph.adj[predecessor][n]))),
                         (n, successor, next(iter(graph.adj[n][successor])))]
            else:
                new_edge = (predecessor, successor)
                edges = [(predecessor, n), (n, successor)]
            merged_line = _get_merging_line(graph, edges[0], edges[1])
            merging_edges = [(new_edge, merged_line)]
            merged_edges[new_edge] = edges.copy()  # copy for non destructive delete in clean merger function
            _clean_merge_mapping(merged_edges, new_edge, edges, True)
        if in_degree == out_degree == 2:
            successors = list(graph.succ[n])
            if all(p in successors for p in graph.pred[n]):
                if successors[1] == successors[0]:
                    continue
                if graph.is_multigraph():
                    back_new_edge = (successors[1], successors[0], graph.new_edge_key(successors[1], successors[0]))
                    forth_new_edge = (successors[0], successors[1], graph.new_edge_key(successors[0], successors[1]))
                    back_edges = [(successors[1], n, next(iter(graph.adj[successors[1]][n]))),
                                  (n, successors[0], next(iter(graph.adj[n][successors[0]])))]
                    forth_edges = [(successors[0], n, next(iter(graph.adj[successors[0]][n]))),
                                   (n, successors[1], next(iter(graph.adj[n][successors[1]])))]
                else:
                    back_new_edge = (successors[1], successors[0])
                    forth_new_edge = (successors[0], successors[1])
                    back_edges = [(successors[1], n), (n, successors[0])]
                    forth_edges = [(successors[0], n), (n, successors[1])]
                back_merged_line = _get_merging_line(graph, back_edges[0], back_edges[1])
                forth_merged_line = _get_merging_line(graph, forth_edges[0], forth_edges[1])
                merging_edges = [(back_new_edge, back_merged_line),
                                 (forth_new_edge, forth_merged_line)]
                merged_edges[back_new_edge] = back_edges.copy()
                _clean_merge_mapping(merged_edges, back_new_edge, back_edges, True)
                merged_edges[forth_new_edge] = forth_edges.copy()
                _clean_merge_mapping(merged_edges, forth_new_edge, forth_edges, True)
        if merging_edges:
            # Remove node (and thus edges)
            graph.remove_node(n)
            # Add merging edges
            for new_edge, line in merging_edges:
                merging_edge_attributes = {}
                if line is not None:
                    merging_edge_attributes[graph.edges_geometry_key] = line
                graph.add_edge(*new_edge, **merging_edge_attributes)
    return merged_edges


def two_degree_node_merge_for_undirected_graphs(graph: GeoGraph, node_filter=no_filter) -> dict:
    """Merge edges that connects two nodes with a unique third node for undirected graphs. Potential nodes to merge are
    nodes with two edges connecting two different nodes. If a replaced edge doesn't have a geometry, the added edge will
    not have a geometry as well.

    Parameters
    ----------
    graph : GeoGraph or GeoMultiGraph
        Graph to modify
    node_filter :
        Evaluates to true if a given node can be merged. (Default value = no_filter)

    Returns
    -------
    dict
        Dictionary indicating for each new edge the merged ones.

    """
    if graph.is_multigraph():
        keys_args = {"keys": True}
    else:
        keys_args = {}
    merged_edges = dict()
    two_degree_nodes = [n for n in graph.nodes() if graph.degree(n) == 2 and node_filter(n)]
    for n in two_degree_nodes:
        edges = list(graph.edges(n, **keys_args))
        assert (len(edges) == 2)
        first_node = edges[0][1] if edges[0][1] != n else edges[0][0]
        second_node = edges[1][1] if edges[1][1] != n else edges[1][0]
        if graph.is_multigraph():
            new_edge = (first_node, second_node, graph.new_edge_key(first_node, second_node))
        else:
            new_edge = (first_node, second_node)
        if first_node == second_node:
            continue
        first_edge_geometry = graph.edges[edges[0]].get(graph.edges_geometry_key, None)
        second_edge_geometry = graph.edges[edges[1]].get(graph.edges_geometry_key, None)
        if first_edge_geometry is not None and second_edge_geometry is not None:
            merged_line = merge_two_lines_with_closest_extremities(first_edge_geometry, second_edge_geometry)
        else:
            merged_line = None
        merged_edges[new_edge] = edges.copy()
        _clean_merge_mapping(merged_edges, new_edge, edges, False)
        # Remove node (and thus edges)
        graph.remove_node(n)
        # Add merging edge
        merging_edge_attributes = {}
        if merged_line is not None:
            merging_edge_attributes[graph.edges_geometry_key] = merged_line
        graph.add_edge(*new_edge, **merging_edge_attributes)
    return merged_edges


def two_degree_node_merge(graph: GeoGraph, node_filter=no_filter) -> dict:
    """Merge edges that connects two nodes with a unique third node.

    Parameters
    ----------
    graph : GeoGraph, GeoDiGraph, GeoMultiGraph or GeoMultiDiGraph
        Graph to modify
    node_filter :
        Evaluates to true if a given node can be merged. (Default value = no_filter)

    Returns
    -------
    dict
        Dictionary indicating for each new edge the merged ones.

    See Also
    --------
    two_degree_node_merge_for_directed_graphs, two_degree_node_merge_for_undirected_graphs

    """
    if graph.is_directed():
        return two_degree_node_merge_for_directed_graphs(graph, node_filter)
    else:
        return two_degree_node_merge_for_undirected_graphs(graph, node_filter)



import geonetworkx as gnx
import osmnx as ox
from shapely.wkt import loads

study_area_wkt = "Polygon ((12.5747751995698902 55.67132899603028307, 12.57548328201693799 55.66792134925386648," \
                 " 12.57951050093452139 55.666947735889174, 12.58490962959326076 55.67053240327735608," \
                 " 12.59194619891079547 55.67424983612435341, 12.60000063674596227 55.67840982050076093," \
                 " 12.60181509801652133 55.68071108845366268, 12.6022576495459262 55.68199448788893591," \
                 " 12.59920404399303351 55.68217150850070141, 12.59513256992251051 55.68274682548892685," \
                 " 12.59291981227548618 55.6826140600301045, 12.58309516832269992 55.67637408346549677," \
                 " 12.5747751995698902 55.67132899603028307))"

study_area = loads(study_area_wkt)

streets_net = ox.graph_from_polygon(study_area)
streets_net = gnx.read_geograph_with_coordinates_attributes(streets_net)
streets_net.name = "streets_net"

ferry_net = ox.graph_from_polygon(study_area, infrastructure='relation["route"="ferry"]', clean_periphery=False,
                                  simplify=False, custom_filter="")

ferry_net = gnx.read_geograph_with_coordinates_attributes(ferry_net)
ferry_net.name = "ferry_net"

merged_graph = gnx.spatial_graph_merge(streets_net, ferry_net)
merged_graph.name = "ferry_and_streets"

gnx.write_geofile(merged_graph, r"D:\projets\GeoNetworkX\GeoNetworkX\geonetworkx\tests\datasets\trash", driver="GeoJSON")
gnx.write_geofile(ferry_net, r"D:\projets\GeoNetworkX\GeoNetworkX\geonetworkx\tests\datasets\trash", driver="GeoJSON")
gnx.write_geofile(streets_net, r"D:\projets\GeoNetworkX\GeoNetworkX\geonetworkx\tests\datasets\trash", driver="GeoJSON")

gdf = ox.footprints_from_polygon(study_area)
gdf["geometry"] = gdf["geometry"].centroid
gdf.drop(columns =["nodes"], inplace=True)

gdf.to_file("copenhagen_buildings.geojson", driver="GeoJSON")
import os
os.getcwd()


# -*- coding: utf-8 -*-
import networkx as nx
import geonetworkx as gnx
import numpy as np
from shapely.geometry import LineString, Point
from nose.tools import assert_true, assert_equal, assert_in
from geonetworkx.utils import crs_equals


ALL_CLASSES = [gnx.GeoGraph, gnx.GeoMultiGraph, gnx.GeoDiGraph, gnx.GeoMultiDiGraph]
SEED = 70595


def assert_almost_intersect(shape1, shape2, msg='', tol=1e-4):
    assert_true(shape1.buffer(tol).intersects(shape2), msg)


def assert_coordinates_almost_equals(coordinates1, coordinates2, msg='', tol=1e-4):
    assert_true(gnx.coordinates_almost_equal(coordinates1, coordinates2, tol), msg)


def assert_lines_almost_equals(line1, line2, msg='', tol=1e-4):
    assert_equal(len(line1.coords), len(line2.coords), msg)
    for c1, c2 in zip(line1.coords, line2.coords):
        assert_coordinates_almost_equals(c1, c2, msg, tol)


def assert_graphs_have_same_edges_geometry(graph1, graph2, msg='', tol=1e-4):
    assert_equal(len(graph1.edges), len(graph2.edges), msg)
    lines1 = nx.get_edge_attributes(graph1, graph1.edges_geometry_key)
    lines2 = nx.get_edge_attributes(graph2, graph2.edges_geometry_key)
    for e in lines1:
        assert_in(e, lines2, msg)
        assert_lines_almost_equals(lines1[e], lines2[e], msg, tol)


def assert_is_subgraph(base_graph, sub_graph, msg=''):
    for n in sub_graph.nodes:
        assert_in(n, base_graph.nodes, msg)
    for e in sub_graph.edges:
        assert_in(e, base_graph.edges, msg)


def assert_graphs_have_same_geonodes(graph1, graph2, msg='', tol=1e-4):
    assert_equal(len(graph1.nodes), len(graph2.nodes), msg)
    coordinates1 = graph1.get_nodes_coordinates()
    coordinates2 = graph1.get_nodes_coordinates()
    assert_equal(len(coordinates1), len(coordinates2), msg)
    for n in coordinates1.keys():
        assert_in(n, coordinates2)
        assert_coordinates_almost_equals(coordinates1[n], coordinates2[n], msg, tol)


def assert_graphs_have_same_spatial_keys(graph1, graph2, msg=''):
    keys1 = graph1.get_spatial_keys()
    keys2 = graph2.get_spatial_keys()
    assert_equal(len(keys1), len(keys2))
    for k in keys1.keys():
        if k != 'crs':
            assert_equal(keys1[k], keys2[k], msg)
        else:
            assert_true(crs_equals(keys1[k], keys2[k]), msg)


def get_random_geograph(nb_nodes):
    global SEED
    edge_creation_prob = 0.1
    g = nx.fast_gnp_random_graph(nb_nodes, edge_creation_prob, seed=SEED, directed=False)
    SEED += 1
    nodes_coords = nx.kamada_kawai_layout(g)
    geometry_attr = "geometry"
    nx.set_node_attributes(g, {n: Point(coords) for n, coords in nodes_coords.items()}, geometry_attr)
    graph = gnx.GeoGraph(g, nodes_geometry_key=geometry_attr, edges_geometry_key=geometry_attr)
    gnx.fill_edges_missing_geometry_attributes(graph)
    return graph


def get_random_geodigraph(nb_nodes):
    global SEED
    edge_creation_prob = 0.1
    g = nx.fast_gnp_random_graph(nb_nodes, edge_creation_prob, seed=SEED, directed=True)
    SEED += 1
    nodes_coords = nx.kamada_kawai_layout(g)
    geometry_attr = "geometry"
    nx.set_node_attributes(g, {n: Point(coords) for n, coords in nodes_coords.items()}, geometry_attr)
    graph = gnx.GeoDiGraph(g, nodes_geometry_key=geometry_attr, edges_geometry_key=geometry_attr)
    gnx.fill_edges_missing_geometry_attributes(graph)
    return graph


def get_random_geomultigraph(nb_nodes):
    global SEED
    edge_creation_prob = 0.1
    g = get_random_geograph(nb_nodes)
    multigraph = gnx.GeoMultiGraph(g)
    np.random.seed(SEED)
    SEED += 1
    for e in list(multigraph.edges):
        if np.random.rand() < edge_creation_prob:
            multigraph.add_edge(e[0], e[1])
    gnx.fill_edges_missing_geometry_attributes(multigraph)
    return multigraph


def get_random_geomultidigraph(nb_nodes):
    global SEED
    edge_creation_prob = 0.1
    g = get_random_geodigraph(nb_nodes)
    multidigraph = gnx.GeoMultiDiGraph(g)
    np.random.seed(SEED)
    SEED += 1
    for e in list(multidigraph.edges):
        if np.random.rand() < edge_creation_prob:
            multidigraph.add_edge(e[0], e[1])
    gnx.fill_edges_missing_geometry_attributes(multidigraph)
    return multidigraph


def get_random_geograph_subclass(nb_nodes, graph_type=gnx.GeoGraph):
    graph_generation_methods = {gnx.GeoGraph: get_random_geograph,
                                gnx.GeoMultiGraph: get_random_geomultigraph,
                                gnx.GeoDiGraph: get_random_geodigraph,
                                gnx.GeoMultiDiGraph: get_random_geomultidigraph}
    g = graph_generation_methods[graph_type](nb_nodes)
    return g


def get_random_geograph_with_wgs84_scale(nb_nodes, graph_type=gnx.GeoGraph):
    g = get_random_geograph_subclass(nb_nodes, graph_type)
    x_func = lambda x: 5.7 + 1e-1 * x
    y_func = lambda y: 45.1 + 1e-1 * y
    transform_point = lambda p: Point([x_func(p.x), y_func(p.y)])
    for n, data in g.nodes(data=True):
        data[g.nodes_geometry_key] = transform_point(data[g.nodes_geometry_key])
    for e in g.edges:
        edge_data = g.edges[e]
        line = edge_data[g.edges_geometry_key]
        line_coords = list(line.coords)
        modified_coords = [(x_func(x), y_func(y)) for x, y in line_coords]
        edge_data[g.edges_geometry_key] = LineString(modified_coords)
    g.crs = gnx.settings.WGS84_CRS
    return g




# -*- coding: utf-8 -*-
from nose.plugins.attrib import attr
from nose.tools import assert_less, assert_not_in, assert_in, assert_equal, assert_true, assert_false
import unittest
import networkx as nx
import geonetworkx as gnx
from geonetworkx.testing.utils import get_random_geograph_subclass, assert_is_subgraph, ALL_CLASSES
import geonetworkx.testing.utils as gnx_tu
from shapely.geometry import Polygon, Point
import math
import numpy as np


NB_VERTICES = 50
gnx_tu.SEED = 70595


@attr("simplify")
class TestSimplify(unittest.TestCase):

    def test_trim_graph_with_polygon(self):
        for method in ["intersects", "within"]:
            for graph_type in ALL_CLASSES:
                g = get_random_geograph_subclass(NB_VERTICES, graph_type)
                nodes_series = g.get_nodes_as_point_series()
                bounds = nodes_series.total_bounds
                bounds[1] = (bounds[1] + bounds[3]) / 2.0  # Taking half bounding box as polygon
                polygon_coordinates = [[bounds[0], bounds[1]],
                                       [bounds[0], bounds[3]],
                                       [bounds[2], bounds[3]],
                                       [bounds[2], bounds[1]]]
                polygon = Polygon(polygon_coordinates)
                new_graph = gnx.trim_graph_with_polygon(g, polygon, copy=True, method=method)
                assert_less(len(new_graph.nodes), len(g.nodes), "Half bounding box trimming must remove some nodes")
                assert_is_subgraph(g, new_graph, "The trimmed graph must be a sub graph of the original graph")

    def test_remove_nan_attributes(self):
        nan_examples = {'t1': None, 't2': math.nan, 't3': np.nan}
        not_nan_examples = {'t4': "abcd", 't5': 1.23, 't6': ["a", "b", "c", None, np.nan, math.nan]}
        for graph_type in ALL_CLASSES:
            for copy in [True, False]:
                with self.subTest(graph_type=graph_type, SEED=gnx_tu.SEED, copy=copy):
                    g = get_random_geograph_subclass(NB_VERTICES, graph_type)
                    for n, d in g.nodes(data=True):
                        d.update(nan_examples)
                        d.update(not_nan_examples)
                    for u, v, d in g.edges(data=True):
                        d.update(nan_examples)
                        d.update(not_nan_examples)
                    if copy:
                        modified_graph = gnx.remove_nan_attributes(g, copy=copy)
                    else:
                        gnx.remove_nan_attributes(g, copy=copy)
                        modified_graph = g

                    for n, d in modified_graph.nodes(data=True):
                        for k in nan_examples.keys():
                            assert_not_in(k, d.keys())
                        for k in not_nan_examples.keys():
                            assert_in(k, d.keys())
                    for u, v, d in modified_graph.edges(data=True):
                        for k in nan_examples.keys():
                            assert_not_in(k, d.keys())
                        for k in not_nan_examples.keys():
                            assert_in(k, d.keys())

    def test_two_degree_merge(self):
        nodes_coordinates = [(1, {"geometry": Point(0, 1)}), (2, {"geometry": Point(1, 1)}),
                             (3, {"geometry": Point(2, 1)}), (4, {"geometry": Point(1, 0)}),
                             (5, {"geometry": Point(1, 2)}), (6, {"geometry": Point(3, 1)}),
                             (7, {"geometry": Point(1, 3)}), (8, {"geometry": Point(1, 4)}),
                             (9, {"geometry": Point(1, 5)}), (10, {"geometry": Point(2, 5)}),
                             (11, {"geometry": Point(3, 5)}), (12, {"geometry": Point(2, 6)}),
                             (13, {"geometry": Point(3, 6)})]
        for graph_type in ALL_CLASSES:
            with self.subTest(graph_type=graph_type):
                g = graph_type()
                g.add_nodes_from(nodes_coordinates)
                nx.add_path(g, [1, 2, 3])
                nx.add_path(g, [4, 2, 5])
                nx.add_path(g, [3, 6])
                nx.add_path(g, [5, 7, 8, 9])
                nx.add_path(g, [11, 10, 9])
                nx.add_path(g, [9, 12, 13])
                if g.is_directed():
                    nx.add_path(g, [6, 3])
                    nx.add_path(g, [13, 12, 9])
                    if g.is_multigraph():
                        g.add_edge(10, 9)
                        merged_nodes = {5, 7, 8, 12}
                        added_edges = {(2, 9), (9, 13), (13, 9)}
                    else:
                        merged_nodes = {5, 7, 8, 10, 12}
                        added_edges = {(2, 9), (11, 9), (9, 13), (13, 9)}
                else:
                    if g.is_multigraph():
                        g.add_edge(10, 9)
                        merged_nodes = {3, 5, 7, 8, 12}
                        added_edges = {(2, 9), (2, 6)}
                    else:
                        merged_nodes = {3, 5, 7, 8, 10, 12}
                        added_edges = {(2, 9), (2, 6), (9, 11), (9, 13)}
                initial_nb_nodes = g.number_of_nodes()
                gnx.fill_edges_missing_geometry_attributes(g)
                if g.is_multigraph():
                    e = (5, 7, 0)
                else:
                    e = (5, 7)
                del g.edges[e][g.edges_geometry_key]
                old_g = g.copy()
                merged_edges = gnx.two_degree_node_merge(g)
                for n in merged_nodes:
                    assert_not_in(n, g.nodes(), "A merged node is a in the simplified graph: %s" % str(n))
                assert_equal(g.number_of_nodes(), initial_nb_nodes - len(merged_nodes))
                for e in added_edges:
                    assert_true(g.has_edge(*e), "A normally added edge is not in the simplified graph: %s" % str(e))
                for new_edge, edges in merged_edges.items():
                    assert_true(g.has_edge(*new_edge), "A new edge is not in the simplified graph: %s" % str(e))
                    for e in edges:
                        assert_false(g.has_edge(*e), "A deleted edge is in the simplified graph: %s" % str(e))
                        assert_true(old_g.has_edge(*e), "A deleted edge is not not in the old graph: %s" % str(e))




# -*- coding: utf-8 -*-
from geonetworkx.testing import get_random_geograph_with_wgs84_scale, get_random_geograph_subclass
from geonetworkx.testing import assert_graphs_have_same_edges_geometry, assert_graphs_have_same_geonodes, ALL_CLASSES, \
                                assert_graphs_have_same_spatial_keys
import geonetworkx.testing.utils as gnx_tu
import geonetworkx as gnx
import geonetworkx.settings as settings
import os
import shutil
from nose.tools import assert_is_instance, assert_equal
from nose.plugins.attrib import attr
import unittest


gnx_tu.SEED = 70595
NB_POINTS = 100


@attr('classes')
class TestClasses(unittest.TestCase):

    def setUp(self):
        file_dir = os.path.dirname(__file__)
        self.results_dir = os.path.join(file_dir, "datasets/results/")
        if not os.path.exists(self.results_dir):
            os.mkdir(self.results_dir)

    def tearDown(self):
        shutil.rmtree(self.results_dir)

    def test_graph_to_directed(self):
        graphs_directed_match = {gnx.GeoGraph: gnx.GeoDiGraph, gnx.GeoMultiGraph: gnx.GeoMultiDiGraph,
                                 gnx.GeoDiGraph: gnx.GeoDiGraph, gnx.GeoMultiDiGraph: gnx.GeoMultiDiGraph}
        for graph_type in ALL_CLASSES:
            with self.subTest(graph_type=graph_type, SEED=gnx_tu.SEED):
                graph = get_random_geograph_subclass(NB_POINTS, graph_type)
                directed_graph = graph.to_directed()
                assert_is_instance(directed_graph, graphs_directed_match[graph_type])

    def test_graph_to_undirected(self):
        graphs_undirected_match = {gnx.GeoGraph: gnx.GeoGraph, gnx.GeoMultiGraph: gnx.GeoMultiGraph,
                                   gnx.GeoDiGraph: gnx.GeoGraph, gnx.GeoMultiDiGraph: gnx.GeoMultiGraph}
        for graph_type in ALL_CLASSES:
            with self.subTest(graph_type=graph_type, SEED=gnx_tu.SEED):
                graph = get_random_geograph_subclass(NB_POINTS, graph_type)
                directed_graph = graph.to_undirected()
                assert_is_instance(directed_graph, graphs_undirected_match[graph_type])

    def test_crs_modification(self):
        graph = get_random_geograph_with_wgs84_scale(NB_POINTS, gnx.GeoMultiDiGraph)
        modified_graph = graph.to_crs(crs={'init': 'epsg:3857'}, inplace=False)
        re_modified_graph = modified_graph.to_crs(crs=gnx.settings.WGS84_CRS, inplace=False)
        assert_graphs_have_same_edges_geometry(graph, re_modified_graph, "Some edge geometries seems to be different"
                                                                         " after re-modification", tol=1e-2)
        assert_graphs_have_same_geonodes(graph, re_modified_graph, "Some nodes seems to be different after "
                                                                   "re-modification")

    def test_nodes_to_gdf(self):
        for graph_type in ALL_CLASSES:
            with self.subTest(graph_type=graph_type, SEED=gnx_tu.SEED):
                g = get_random_geograph_subclass(NB_POINTS, graph_type)
                gdf = g.nodes_to_gdf()
                assert_equal(g.number_of_nodes(), len(gdf))

    def test_edges_to_gdf(self):
        for graph_type in ALL_CLASSES:
            with self.subTest(graph_type=graph_type, SEED=gnx_tu.SEED):
                g = get_random_geograph_subclass(NB_POINTS, graph_type)
                gdf = g.edges_to_gdf()
                assert_equal(g.number_of_edges(), len(gdf))

    def test_spatial_keys_persistence(self):
        for graph_type in ALL_CLASSES:
            with self.subTest(graph_type=graph_type, SEED=gnx_tu.SEED):
                g = get_random_geograph_subclass(NB_POINTS, graph_type)
                gnx.rename_edges_attribute(g, g.edges_geometry_key, "abcd")
                g.edges_geometry_key = "abcd"
                gnx.rename_nodes_attribute(g, g.nodes_geometry_key, "efgh")
                g.nodes_geometry_key = "efgh"
                g.crs = {'init': 'epsg:3945'}
                g2 = g.copy(as_view=False)
                assert_graphs_have_same_spatial_keys(g, g2)
                g3 = g.to_undirected(as_view=False)
                assert_graphs_have_same_spatial_keys(g, g3)
                g4 = g.to_directed(as_view=False)
                assert_graphs_have_same_spatial_keys(g, g4)

    def test_add_edges_gdf(self):
        for graph_type in ALL_CLASSES:
            with self.subTest(graph_type=graph_type, SEED=gnx_tu.SEED):
                g = get_random_geograph_subclass(NB_POINTS, graph_type)
                initial_edges = list(g.edges)
                nb_initial_edges = len(initial_edges)
                g2 = get_random_geograph_subclass(NB_POINTS, graph_type)
                edges_gdf = g2.edges_to_gdf()
                g.add_edges_from_gdf(edges_gdf, settings.EDGE_FIRST_NODE_COLUMN_NAME,
                                     gnx.settings.EDGE_SECOND_NODE_COLUMN_NAME)
                if g.is_multigraph():
                    g2_nb_edges = g2.number_of_edges()
                    assert_equal(g.number_of_edges(), g2_nb_edges + nb_initial_edges,
                                 "Resulting graph have not the right number of edges.")
                else:
                    g2_nb_distinct_edges = len([e for e in g2.edges if e not in initial_edges])
                    assert_equal(g.number_of_edges(), g2_nb_distinct_edges + nb_initial_edges,
                                 "Resulting graph have not the right number of edges.")



# -*- coding: utf-8 -*-
import os
import pickle
import numpy as np
import networkx as nx
import geopandas as gpd
from shapely.geometry import Point, LineString
from shapely.wkt import loads
from .geograph import GeoGraph
from .geomultigraph import GeoMultiGraph
from .geodigraph import GeoDiGraph
from .geomultidigraph import GeoMultiDiGraph
import geonetworkx.settings as settings
from geonetworkx.utils import get_crs_as_str


def parse_graph_as_geograph(graph, **attr):
    """Parse a ``networkx.Graph`` as a ``geonetworkx.GeoGraph`` with the closest geonetworkx graph type.

    Parameters
    ----------
    graph : nx.Graph, nx.DiGraph, nx.MultiGraph or nx.MultiDiGraph
        
    **attr :
        Potential spatial keys.

    Returns
    -------
    GeoGraph, GeoDiGraph, GeoMultiGraph or GeoMultiDiGraph
        Depending the orientation and multi edges properties.
    """
    if graph.is_directed():
        if graph.is_multigraph():
            geograph = GeoMultiDiGraph(graph, **attr)
        else:
            geograph = GeoDiGraph(graph, **attr)
    else:
        if graph.is_multigraph():
            geograph = GeoMultiGraph(graph, **attr)
        else:
            geograph = GeoGraph(graph, **attr)
    return geograph


def get_graph_with_wkt_geometry(geograph: GeoGraph) -> nx.Graph:
    """Modify the edges geometry attribute to a well-known text format to make the graph writable is some text formats.
    The returned graph is not as operational as the given one (edge geometries has been removed).

    Parameters
    ----------
    geograph: GeoGraph :
        Geograph to transform

    Returns
    -------
    nx.Graph
        A networkx graph with WKT geometries instead of shapely objects.

    See Also
    --------
    parse_nodes_attribute_as_wkt

    """
    graph_shallow_copy = geograph.to_nx_class()(geograph)
    nodes_geometries = nx.get_node_attributes(graph_shallow_copy, geograph.nodes_geometry_key)
    for n, point in nodes_geometries.items():
        if hasattr(point, "wkt"):
            graph_shallow_copy.nodes[n][geograph.nodes_geometry_key] = point.wkt
    edge_geometries = nx.get_edge_attributes(graph_shallow_copy, geograph.edges_geometry_key)
    for e, line in edge_geometries.items():
        if hasattr(line, "wkt"):
            graph_shallow_copy.edges[e][geograph.edges_geometry_key] = line.wkt
    return graph_shallow_copy


def parse_nodes_attribute_as_wkt(graph: nx.Graph, attribute_name: str):
    """Transform a graph nodes attribute from WKT to shapely objects. Attribute is replaced.

    Parameters
    ----------
    graph: nx.Graph :
        Graph to modify and parse
    attribute_name: str :
        Attribute to parse the nodes geometries

    See Also
    --------
    get_graph_with_wkt_geometry, parse_edges_attribute_as_wkt

    """
    wkt_points = nx.get_node_attributes(graph, attribute_name)
    for n, w in wkt_points.items():
        graph.nodes[n][attribute_name] = loads(w)


def parse_edges_attribute_as_wkt(graph: nx.Graph, attribute_name: str):
    """Transform a graph edges attribute from WKT to shapely objects. Attribute is replaced.

    Parameters
    ----------
    graph: nx.Graph :
        Graph to modify and parse
    attribute_name: str :
        Attribute to parse the edges geometries

    See Also
    --------
    get_graph_with_wkt_geometry, parse_nodes_attribute_as_wkt

    """
    wkt_lines = nx.get_edge_attributes(graph, attribute_name)
    for e, w in wkt_lines.items():
        graph.edges[e][attribute_name] = loads(w)


def stringify_crs(graph: GeoGraph):
    """Write the CRS attribute as a string."""
    if 'crs' in graph.graph and graph.graph['crs'] is not None:
        if not isinstance(graph.graph['crs'], str):
            graph.graph['crs'] = get_crs_as_str(graph.crs)


def read_gpickle(path, **attr):
    """Read geograph object in Python pickle format.

    Parameters
    ----------
    path : str
        Path where to read a graph object.
    **attr :
        Potential spatial keys.

    Returns
    -------
    GeoGraph, GeoDiGraph, GeoMultiGraph or GeoMultiDiGraph
        The parsed geograph.

    See Also
    --------
    write_gpickle, nx.read_gpickle, nx.write_gpickle

    """
    graph = nx.read_gpickle(path)
    if type(graph) in [nx.Graph, nx.DiGraph, nx.MultiDiGraph, nx.MultiGraph]:
        return parse_graph_as_geograph(graph, **attr)
    else:
        return graph


def write_gpickle(geograph, path, protocol=pickle.HIGHEST_PROTOCOL):
    """Write geograph object in Python pickle format.

    Parameters
    ----------
    geograph : GeoGraph, GeoDiGraph, GeoMultiGraph, GeoMultiDiGraph
        Geograph to write
    path :
        Path where to right the pickle file.
    protocol :
         See pickle protocols (Default value = pickle.HIGHEST_PROTOCOL).

    Returns
    -------

    See Also
    --------
    read_gpickle, nx.read_gpickle, nx.write_gpickle

    """
    nx.write_gpickle(geograph, path, protocol)


def read_graphml(path, node_type=str, edge_key_type=int, **attr) -> GeoGraph:
    """Read graph in GraphML format from path.

    Parameters
    ----------
    path :
        File path to the graphml file.
    node_type :
        See ``nx.read_graphml`` (Default value = str)
    edge_key_type :
        See ``nx.read_graphml`` (Default value = int)
    **attr :
        Potential spatial keys

    Returns
    -------
    GeoGraph, GeoDiGraph, GeoMultiGraph, GeoMultiDiGraph
        Parsed Geograph

    See Also
    --------
    write_graphml, nx.read_graphml, nx.write_graphml

    """
    graph = nx.read_graphml(path, node_type, edge_key_type)
    nodes_geometry_attr = attr.get("nodes_geometry_key",
                                   graph.graph.get("nodes_geometry_key",
                                                   settings.NODES_GEOMETRY_DEFAULT_KEY))
    parse_nodes_attribute_as_wkt(graph, nodes_geometry_attr)
    edges_geometry_attr = attr.get("edges_geometry_key",
                                   graph.graph.get("edges_geometry_key",
                                                   settings.EDGES_GEOMETRY_DEFAULT_KEY))
    parse_edges_attribute_as_wkt(graph, edges_geometry_attr)
    return parse_graph_as_geograph(graph, **attr)


def write_graphml(geograph, path, encoding='utf-8', prettyprint=True, infer_numeric_types=False):
    """Generate GraphML file for the given geograph.

    Parameters
    ----------
    geograph : GeoGraph, GeoDiGraph, GeoMultiGraph, GeoMultiDiGraph
        Geograph to write
    path : str
        Writing file path
    encoding :
         See ``nx.write_graphml`` (Default value = 'utf-8')
    prettyprint :
         See ``nx.write_graphml`` (Default value = True)
    infer_numeric_types :
         See ``nx.write_graphml`` (Default value = False)

    See Also
    --------
    read_graphml, nx.read_graphml, nx.write_graphml

    """
    graph_wkt = get_graph_with_wkt_geometry(geograph)
    stringify_crs(graph_wkt)
    nx.write_graphml(graph_wkt, path, encoding, prettyprint, infer_numeric_types)


def graph_nodes_to_gdf(graph: GeoGraph) -> gpd.GeoDataFrame:
    """Create and fill a GeoDataFrame (geopandas) from nodes of a networkX graph. The ``'geometry'`` attribute is used for
    shapes.

    Parameters
    ----------
    graph : GeoGraph
        Graph to parse

    Returns
    -------
    gpd.GeoDataFrame
        The resulting GeoDataFrame : one row is a node

    """
    nodes = {node: data for node, data in graph.nodes(data=True)}
    gdf_nodes = gpd.GeoDataFrame(nodes).T
    gdf_nodes['id'] = gdf_nodes.index
    gdf_nodes['geometry'] = gdf_nodes.apply(lambda row: Point(row['x'], row['y']), axis=1)
    if 'crs' in graph.graph:
        gdf_nodes.crs = graph.graph['crs']
    if 'name' in graph.graph:
        gdf_nodes.gdf_name = '{}_nodes'.format(graph.graph['name'])
    return gdf_nodes


def graph_edges_to_gdf(graph: nx.Graph) -> gpd.GeoDataFrame:
    """Create and fill a GeoDataFrame (geopandas) from edges of a networkX graph. The ``'geometry'`` attribute is used
     for shapes.

    Parameters
    ----------
    graph : nx.Graph
        Graph to parse

    Returns
    -------
    gpd.GeoDataFrame
        The resulting GeoDataFrame : one row is an edge

    """
    # create a list to hold our edges, then loop through each edge in the graph
    edges = []
    for u, v, data in graph.edges(data=True):
        # for each edge, add key and all attributes in data dict to the edge_details
        edge_details = {'u': u, 'v': v}
        for attr_key in data:
            edge_details[attr_key] = data[attr_key]
        # if edge doesn't already have a geometry attribute, create one now
        if 'geometry' not in data:
            point_u = Point((graph.nodes[u]['x'], graph.nodes[u]['y']))
            point_v = Point((graph.nodes[v]['x'], graph.nodes[v]['y']))
            edge_details['geometry'] = LineString([point_u, point_v])
        edges.append(edge_details)
    # create a GeoDataFrame from the list of edges and set the CRS
    gdf_edges = gpd.GeoDataFrame(edges)
    if 'crs' in graph.graph:
        gdf_edges.crs = graph.graph['crs']
    if 'name' in graph.graph:
        gdf_edges.gdf_name = '{}_edges'.format(graph.graph['name'])
    return gdf_edges


def parse_bool_columns_as_int(gdf: gpd.GeoDataFrame):
    """Transform bool columns into integer columns.

    Parameters
    ----------
    gdf: gpd.GeoDataFrame :
        GeoDataFrame to modify
    """
    for c in gdf.columns:
        if gdf[c].dtype == "bool":
            gdf[c] = gdf[c].astype("int")


def parse_numpy_types(gdf: gpd.GeoDataFrame):
    """Transform numpy types as scalar types.

    Parameters
    ----------
    gdf: gpd.GeoDataFrame :
        GeoDataFrame to modify
    """
    for c in gdf.columns:
        if any(map(lambda x: isinstance(x, np.generic), gdf[c])):
            for i in gdf.index:
                if isinstance(gdf.loc[i, c], np.generic):
                    gdf.loc[i, c] = np.asscalar(gdf.loc[i, c])


def stringify_unwritable_columns(gdf: gpd.GeoDataFrame):
    """Transform elements which have type bool or list to string

    Parameters
    ----------
    gdf: gpd.GeoDataFrame :
        GeoDataFrame to modify
    """
    valid_columns_types = ("int64", "float64")
    for c in gdf.columns:
        if not gdf[c].dtype in valid_columns_types and c != gdf._geometry_column_name:
            gdf[c] = list(map(str, gdf[c]))


def cast_for_fiona(gdf: gpd.GeoDataFrame):
    """Transform elements so that attributes can be writable by fiona.

    Parameters
    ----------
    gdf: gpd.GeoDataFrame :
        GeoDataFrame to modify
    """
    parse_bool_columns_as_int(gdf)
    parse_numpy_types(gdf)
    stringify_unwritable_columns(gdf)


def write_edges_to_geofile(graph: GeoGraph, file_name, driver="GPKG", fiona_cast=True):
    """Writes the edges of a geograph as a geographic file.

    Parameters
    ----------
    graph : GeoGraph, GeoDiGraph, GeoMultiGraph, GeoMultiDiGraph
        Graph to export
    file_name :
        File name (with path)
    driver :
        driver for export file format (GPKG, geojson, etc: can be found from ``fiona.supported_drivers``)
        (Default value = "GPKG")
    fiona_cast :
        If true, methods for casting types to writable fiona types are used (Default value = True)

    See Also
    --------
    write_geofile, write_nodes_to_geofile

    """
    gdf_edges = graph.edges_to_gdf()
    if fiona_cast:
        cast_for_fiona(gdf_edges)
    gdf_edges.to_file(file_name, driver=driver)


def write_nodes_to_geofile(graph: GeoGraph, file_name, driver="GPKG", fiona_cast=True):
    """Writes the nodes of a geograph as a geographic file.

    Parameters
    ----------
    graph : GeoGraph, GeoDiGraph, GeoMultiGraph, GeoMultiDiGraph
        Graph to export
    file_name :
        File name (with path)
    driver :
        driver for export file format (GPKG, geojson, etc: can be found from ``fiona.supported_drivers``)
        (Default value = "GPKG")
    fiona_cast :
        If true, methods for casting types to writable fiona types are used (Default value = True)

    See Also
    --------
    write_geofile, write_edges_to_geofile

    """
    gdf_nodes = graph.nodes_to_gdf()
    if fiona_cast:
        cast_for_fiona(gdf_nodes)
    gdf_nodes.to_file(file_name, driver=driver)


def write_geofile(graph: GeoGraph, path='./', nodes=True, edges=True, driver="GPKG", fiona_cast=True):
    """Export a networkx graph as a geographic files. Two files are generated: one for the nodes and one for the edges.
    The files names will be prefixed by the graph name and suffixed by "_edges" or "_nodes".

    Parameters
    ----------
    graph :
        Graph to export
    path :
        export directory (Default value = './')
    nodes :
        boolean to indicate whether export nodes or not. (Default value = True)
    edges :
        boolean to indicate whether export edges or not. (Default value = True)
    driver :
        driver for export file format (GPKG, geojson, etc: can be found from ``fiona.supported_drivers``)
         (Default value = "GPKG")
    fiona_cast :
        If true, methods for casting types to writable fiona types are used (Default value = True)

    See Also
    --------
    write_nodes_to_geofile, write_edges_to_geofile

    """
    if not os.path.exists(path):
        os.mkdir(path)
    if nodes:
        file_name = os.path.join(path, '{}_nodes'.format(graph.name))
        file_name += settings.KNOWN_FILES_EXTENSION.get(driver, "")
        write_nodes_to_geofile(graph, file_name, driver, fiona_cast)
    if edges:
        file_name = os.path.join(path, '{}_edges'.format(graph.name))
        file_name += settings.KNOWN_FILES_EXTENSION.get(driver, "")
        write_edges_to_geofile(graph, file_name, driver, fiona_cast)


def read_geograph_with_coordinates_attributes(graph: nx.Graph, x_key='x', y_key='y', **attr) -> GeoGraph:
    """Parse a `networkx` graph which have node's coordinates as attribute. This method can be useful to parse an output
    graph of the `osmnx` package.

    Parameters
    ----------
    graph : nx.Graph
        Given graph to parse. All nodes must have the ``x_key`` and ``y_key`` attributes.
    x_key :
        x-coordinates attribute to parse (Default value = 'x')
    y_key :
        y-coordinates attribute to parse (Default value = 'y')
    **attr :
        Optional geograph spatial keys.

    Returns
    -------
    GeoGraph, GeoDiGraph, GeoMultiGraph, GeoMultiDiGraph
        The parsed geograph (shallow copy of the input graph).

    """
    graph = graph.__class__(graph)
    x_coords = nx.get_node_attributes(graph, x_key)
    y_coords = nx.get_node_attributes(graph, y_key)
    nodes_geometry_key = attr.pop("nodes_geometry_key", settings.NODES_GEOMETRY_DEFAULT_KEY)
    edges_geometry_key = attr.pop("edges_geometry_key", settings.EDGES_GEOMETRY_DEFAULT_KEY)
    for n in graph.nodes:
        if n not in x_coords or n not in y_coords:
            raise ValueError("Unable to find coordinates for node : '%s'" % str(n))
        point = Point([x_coords[n], y_coords[n]])
        graph.nodes[n][nodes_geometry_key] = point
    graph.nodes_geometry_key = nodes_geometry_key
    graph.edges_geometry_key = edges_geometry_key
    return parse_graph_as_geograph(graph, **attr)


def read_geofiles(nodes_file_path: str, edges_file_path: str,
                  directed=True, multigraph=False,
                  node_index_attr=settings.NODE_ID_COLUMN_NAME,
                  edge_first_node_attr=settings.EDGE_FIRST_NODE_COLUMN_NAME,
                  edge_second_node_attr=settings.EDGE_SECOND_NODE_COLUMN_NAME):
    """Read geofiles to create a ``GeoGraph``. Geofiles are read with ``geopandas.read_file`` method.

    Parameters
    ----------
    nodes_file_path : str
        File path of nodes.
    edges_file_path : str
        File path of edges.
    directed : bool
        If ``True``, returns a directed graph. (Default value = True)
    multigraph : bool
        If ``True``, returns a multigraph. (Default value = False)
    node_index_attr : str
        Node id attribute in the geofile for nodes labeling. (Default value = settings.NODE_ID_COLUMN_NAME)
    edge_first_node_attr : str
        Edge first node attribute in the geofile. (Default value = settings.EDGE_FIRST_NODE_COLUMN_NAME)
    edge_second_node_attr : str
        Edge second node attribute in the geofile. (Default value = settings.EDGE_SECOND_NODE_COLUMN_NAME)

    Returns
    -------
    GeoGraph, GeoDiGraph, GeoMultiGraph, GeoMultiDiGraph
        A parsed ``Geograph``.

    See Also
    --------
    GeoGraph.add_nodes_from_gdf, GeoGraph.add_edges_from_gdf, geopandas.read_file

    """
    if directed:
        if multigraph:
            graph = GeoMultiDiGraph()
        else:
            graph = GeoDiGraph()
    else:
        if multigraph:
            graph = GeoMultiGraph()
        else:
            graph = GeoGraph()
    if nodes_file_path is not None:
        nodes_gdf = gpd.read_file(nodes_file_path)
        graph.nodes_geometry_key = nodes_gdf._geometry_column_name
        graph.crs = nodes_gdf.crs
        graph.add_nodes_from_gdf(nodes_gdf, node_index_attr)
    if edges_file_path is not None:
        edges_gdf = gpd.read_file(edges_file_path)
        if graph.crs is None and edges_gdf.crs is not None:
            graph.crs = edges_gdf.crs
        graph.add_edges_from_gdf(edges_gdf, edge_first_node_attr, edge_second_node_attr)
    return graph



# -*- coding: utf-8 -*-
import numpy as np
import networkx as nx
from networkx.classes.filters import no_filter
import geopandas as gpd
from shapely.geometry import LineString
from geonetworkx.geograph import GeoGraph
import geonetworkx.settings as settings
from geonetworkx.geometry_operations import get_closest_line_from_points, split_line, coordinates_almost_equal
from geonetworkx.utils import get_new_node_unique_name, euclidian_distance, get_line_ordered_edge, is_nan, compose
from collections import defaultdict


def spatial_points_merge(graph: GeoGraph, points_gdf: gpd.GeoDataFrame, inplace=False, merge_direction="both",
                         node_filter=no_filter, edge_filter=no_filter, intersection_nodes_attr=None) -> GeoGraph:
    """Merge given points as node with a spatial merge. Points are projected on the closest edge of the
    graph and an intersection node is added if necessary. If two nodes a given point and a node have the same name, with
    equal coordinates, then the node is considered as already in the graph. A discretization tolerance
    (``settings.DISCRETIZATION_TOLERANCE``) is used for edges lines and is set by default to a constant matching the
    WGS84 crs. If another crs is used, results may be inconsistent (high computational time or inaccuracy). New nodes
    created from the geodataframe have attributes described by other columns (except if an attribute value is `nan`).
    When a point is projected on an edge, this edge is removed and replaced by two others that connect the extremities
    to the intersection node. A reference to the original edge is kept on theses new edges with the attribute
    ``settings.ORIGINAL_EDGE_KEY``. The original edge is the oldest parent of the new edge, to have the direct parent,
    the attribute has to be cleant first.

    Parameters
    ----------
    graph : GeoGraph, GeoDiGraph, GeoMultiGraph or GeoMultiDiGraph
        A GeoGraph or derived class describing a spatial graph.
    points_gdf : gpd.GeoDataFrame
        A list of point describing new nodes to add.
    inplace : bool
        If True, do operation inplace and return None. (Default value = False)
    merge_direction : str
        For directed graphs only:
        
        * ``'both'``: 2 edges are added: graph -> new node and new node -> graph
        * ``'in'``: 1 edge is added: new_node -> graph
        * ``'out'``: 1 edge is added: graph -> new_node (Default value = "both")
    node_filter :
        A node filter (lambda) to exclude nodes (and by the way all concerned edges) from the projection
        operation. (Default value = no_filter)
    edge_filter :
        An edge filter (lambda) to exclude edges on which the projection will not take place. (Default value = no_filter)
    intersection_nodes_attr : str
        A dictionary of attributes (constant for all added intersection nodes). (Default value = None)

    Returns
    -------
    None or GeoGraph
        If not inplace, the created graph.


    See Also
    --------
    spatial_graph_merge

    """
    if not inplace:
        graph = graph.copy()
    # 1. Find closest edge for each point
    graph_view = nx.graphviews.subgraph_view(graph, filter_node=node_filter, filter_edge=edge_filter)
    edges_as_lines = nx.get_edge_attributes(graph_view, graph.edges_geometry_key)
    if len(edges_as_lines) == 0:
        raise ValueError("No edge geometry has been found in the given merging edges, at least one edge geometry is"
                         " required for a merge operation")
    points = points_gdf.geometry
    points_coords = np.array([[p.x, p.y] for p in points])
    lines_indexes = get_closest_line_from_points(points_coords, edges_as_lines.values())
    edges_to_split = defaultdict(dict)
    # Add node, intersection node and edge (node, intersection node)
    for p, p_index, point in zip(range(len(points_gdf)), points_gdf.index, points):
        # 1.1 Add given node
        if p_index in graph.nodes:
            if coordinates_almost_equal([point.x, point.y], graph.get_node_coordinates(p_index)):
                continue
            else:
                node_name = get_new_node_unique_name(graph, p_index)
        else:
            node_name = p_index
        node_info = {c: points_gdf.at[p_index, c] for c in points_gdf.columns if not is_nan(points_gdf.at[p_index, c])}
        node_info[graph.nodes_geometry_key] = point
        graph.add_node(node_name, **node_info)
        # 1.2 Add projected node if necessary
        closest_edge_name = list(edges_as_lines.keys())[lines_indexes[p]]
        closest_line = edges_as_lines[closest_edge_name]
        closest_line_length = closest_line.length
        intersection_distance_on_line = closest_line.project(point)
        # if the intersection point is on the edge
        if 0 < intersection_distance_on_line < closest_line_length:
            projected_point = closest_line.interpolate(intersection_distance_on_line)
            intersection_node_name = get_new_node_unique_name(graph, settings.INTERSECTION_PREFIX + str(p_index))
            intersection_node_info = {graph.nodes_geometry_key: projected_point}
            if intersection_nodes_attr is not None:
                intersection_node_info.update(intersection_nodes_attr)
            graph.add_node(intersection_node_name, **intersection_node_info)
            # Store line to modify
            edges_to_split[closest_edge_name][intersection_node_name] = intersection_distance_on_line
        else:  # if the intersection point is on of the two edge extremities
            first_node = closest_edge_name[0]
            first_node_point = graph.nodes[first_node][graph.nodes_geometry_key]
            second_node = closest_edge_name[1]
            second_node_point = graph.nodes[second_node][graph.nodes_geometry_key]
            distance_to_first_extremity = euclidian_distance(point, first_node_point)
            distance_to_second_extremity = euclidian_distance(point, second_node_point)
            if distance_to_first_extremity < distance_to_second_extremity:
                intersection_node_name = first_node
            else:
                intersection_node_name = second_node
        # 1.3 Add edge : node <-> intersection_node
        in_edge_data = {graph.edges_geometry_key: LineString([graph.get_node_coordinates(node_name),
                                                              graph.get_node_coordinates(intersection_node_name)])}
        if graph.is_directed():
            out_edge_data = {graph.edges_geometry_key: LineString([graph.get_node_coordinates(intersection_node_name),
                                                                   graph.get_node_coordinates(node_name)])}
            if merge_direction == "both":
                graph.add_edge(node_name, intersection_node_name, **in_edge_data)
                graph.add_edge(intersection_node_name, node_name, **out_edge_data)
            elif merge_direction == "in":
                graph.add_edge(node_name, intersection_node_name, **in_edge_data)
            else:  # "out"
                graph.add_edge(intersection_node_name, node_name, **out_edge_data)
        else:
            graph.add_edge(node_name, intersection_node_name, **in_edge_data)
    # 2. Split edges where a node have been projected
    for e in edges_to_split:
        intersection_nodes = edges_to_split[e]
        if len(intersection_nodes) > 0:
            initial_line = edges_as_lines[e]
            # 2.1 remove initial edge and keep in memory the original edge
            original_edge_data = {settings.ORIGINAL_EDGE_KEY: graph.edges[e].get(settings.ORIGINAL_EDGE_KEY, e)}
            if graph.has_edge(*e):
                graph.remove_edge(*e)
            # 2.2 cut the initial line
            sorted_intersection_nodes = sorted(intersection_nodes.keys(), key=lambda n: intersection_nodes[n])
            distances_on_initial_line = [intersection_nodes[n] for n in sorted_intersection_nodes]
            split_lines = []
            cut_lines = split_line(initial_line, distances_on_initial_line[0])
            split_lines.append(cut_lines[0])
            for i in range(len(sorted_intersection_nodes) - 1):
                cut_lines = split_line(cut_lines[1], distances_on_initial_line[i + 1] - distances_on_initial_line[i])
                split_lines.append(cut_lines[0])
            split_lines.append(cut_lines[1])
            # 2.2 add intermediary edges
            oriented_edge = get_line_ordered_edge(graph, e, initial_line)
            first_edge_data = {graph.edges_geometry_key: split_lines[0], **original_edge_data}
            graph.add_edge(oriented_edge[0], sorted_intersection_nodes[0], **first_edge_data)
            last_edge_data = {graph.edges_geometry_key: split_lines[-1], **original_edge_data}
            graph.add_edge(sorted_intersection_nodes[-1], oriented_edge[1], **last_edge_data)
            for i in range(len(sorted_intersection_nodes) - 1):
                edge_data = {graph.edges_geometry_key: split_lines[i + 1], **original_edge_data}
                graph.add_edge(sorted_intersection_nodes[i], sorted_intersection_nodes[i + 1], **edge_data)
    if not inplace:
        return graph


def spatial_graph_merge(base_graph: GeoGraph, other_graph: GeoGraph,
                        inplace=False, merge_direction="both", node_filter=None, intersection_nodes_attr=None):
    """Operates spatial merge between two graphs. Spatial edge projection is used on merging nodes (see
    ``spatial_points_merge``). The ``base_graph`` attributes have higher priority than the ``other_graph`` attributes (
    i.e. if graphs have common graph attributes, nodes or edges, the ``base_graph`` attributes will be kept).

    Parameters
    ----------
    base_graph : GeoGraph, GeoDiGraph, GeoMultiGraph or GeoMultiDiGraph
        Base graph on which the merge operation is done.
    other_graph : GeoGraph, GeoDiGraph, GeoMultiGraph or GeoMultiDiGraph
        Input graph to merge. Modified graph if operation is done inplace.
    inplace : bool
        If True, do operation inplace and return None. (Default value = False)
    merge_direction : str
        See ``spatial_points_merge`` (Default value = "both")
    node_filter :
        Lambda returning if a given node (from the ``other_graph`` graph) has to be merged. (Default value = None)
    intersection_nodes_attr : str
        A dictionary of attributes (constant for all added intersection nodes). (Default value = None)


    Returns
    -------
    None or GeoGraph
        A new graph with the same type as ``base_graph`` if not inplace.

    See Also
    --------
    spatial_points_merge

    """
    if base_graph.is_directed() != other_graph.is_directed():
        raise ValueError("Merging a directed graph and an undirected graph is ambiguous")
    if base_graph.is_multigraph() != other_graph.is_multigraph():
        raise ValueError("Given graphs must both be graphs or multigraph")
    if node_filter is not None:
        other_graph_view = nx.graphviews.subgraph_view(other_graph, node_filter)
    else:
        other_graph_view = other_graph
    if other_graph_view.number_of_nodes() == 0:
        raise ValueError("Given graph has no nodes to project.")
    nodes_gdf = other_graph_view.nodes_to_gdf()
    if inplace:
        spatial_points_merge(base_graph, nodes_gdf, inplace=inplace, merge_direction=merge_direction,
                             intersection_nodes_attr=intersection_nodes_attr)
        merged_graph = base_graph
    else:
        merged_graph = spatial_points_merge(base_graph, nodes_gdf, inplace=inplace, merge_direction=merge_direction,
                                            intersection_nodes_attr=intersection_nodes_attr)
    merged_graph = compose(other_graph, merged_graph)
    if not inplace:
        return merged_graph
